import sys

from osgeo import osr

from buzzard import _tools

class ASource(object):
    """Base abstract class defining the common behavior of all sources opened in the Dataset.

    Features Defined
    ----------------
    - Has a `stored` spatial reference
    - Has a `virtual` spatial reference that is influenced by the Dataset's opening mode
    - Can be closed
    """

    def __init__(self, ds, back):
        self._ds = ds
        self._back = back

    @property
    def wkt_stored(self):
        """The spatial reference that can be found in the metadata of a source, in wkt format.

        string or None
        """
        return self._back.wkt_stored

    @property
    def proj4_stored(self):
        """The spatial reference that can be found in the metadata of a source, in proj4 format.

        string or None
        """
        return self._back.proj4_stored

    @property
    def wkt_virtual(self):
        """The spatial reference considered to be written in the metadata of a source, in wkt
        format.

        string or None
        """
        return self._back.wkt_virtual

    @property
    def proj4_virtual(self):
        """The spatial reference considered to be written in the metadata of a source, in proj4
        format.

        string or None
        """
        return self._back.proj4_virtual

    def get_keys(self):
        """Get the list of keys under which this source is registered to in the Dataset"""
        return list(self._ds._keys_of_source[self])

    @property
    def close(self):
        """Close a source with a call or a context management.
        The `close` attribute returns an object that can be both called and used in a with statement

        Examples
        --------
        >>> ds.dem.close()
        >>> with ds.dem.close:
                # code...
        >>> with ds.acreate_raster('result.tif', fp, float, 1).close as result:
                # code...
        >>> with ds.acreate_vector('results.shp', 'linestring').close as roofs:
                # code...
        """
        def _close():
            self._back.close()
            self._ds._unregister(self)
            del self._ds
            del self._back

        return _CloseRoutine(self, _close)

    def __del__(self):
        if hasattr(self, '_ds'):
            self.close()

    # Deprecation
    wkt_origin = _tools.deprecation_pool.wrap_property(
        'wkt_virtual',
        '0.4.4'
    )
    proj4_origin = _tools.deprecation_pool.wrap_property(
        'proj4_virtual',
        '0.4.4'
    )

class ABackSource(object):
    """Implementation of ASource's specifications"""

    def __init__(self, back_ds, wkt_stored, rect, **kwargs):
        wkt_virtual = back_ds.virtual_of_stored_given_mode(
            wkt_stored, back_ds.wkt_work, back_ds.wkt_fallback, back_ds.wkt_forced,
        )

        if wkt_virtual is not None:
            sr_virtual = osr.SpatialReference(wkt_virtual)
        else:
            sr_virtual = None

        to_work, to_virtual = back_ds.get_transforms(sr_virtual, rect)

        self.back_ds = back_ds
        self.wkt_stored = wkt_stored
        self.wkt_virtual = wkt_virtual
        self.to_work = to_work
        self.to_virtual = to_virtual
        super().__init__(**kwargs)

    def close(self):
        """Virtual method:
        - May be overriden
        - Should always be called
        """
        del self.back_ds

    @property
    def proj4_virtual(self):
        if self.wkt_virtual is None:
            return None # pragma: no cover
        return osr.SpatialReference(self.wkt_virtual).ExportToProj4()

    @property
    def proj4_stored(self):
        if self.wkt_stored is None:
            return None # pragma: no cover
        return osr.SpatialReference(self.wkt_stored).ExportToProj4()

_CloseRoutine = type('_CloseRoutine', (_tools.CallOrContext,), {
    '__doc__': ASource.close.__doc__,
})

if sys.version_info < (3, 6):
    # https://www.python.org/dev/peps/pep-0487/
    for k, v in ASource.__dict__.items():
        if hasattr(v, '__set_name__'):
            v.__set_name__(ASource, k)



# pylint: disable=redefined-outer-name, unused-argument

from __future__ import division, print_function

import uuid
import os
import sys
import multiprocessing as mp
import multiprocessing.pool

import numpy as np
import pytest
import shapely.geometry as sg

import buzzard as buzz
from buzzard.test.tools import fpeq


def test_vector():
    ds = buzz.Dataset(max_active=2)
    meta = dict(
        type='point',
    )

    def statuses(*args):
        l = tuple([
            (
                ds._back.idle_count(prox._back.uid),
                ds._back.used_count(prox._back.uid),
                prox.active_count,
                prox.active
            )
            for prox in args
        ])
        return l

    assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (0, 0, 0)
    with ds.acreate_vector('/tmp/v1.shp', **meta).delete as r1:
        r1.insert_data([0, 0])

        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (1, 0, 1)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (1, 0, 1, True)

        # Iteration 1 - exn
        it = r1.iter_data()
        next(it)
        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (0, 1, 1)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (0, 1, 1, True)

        try:
            next(it)
        except StopIteration:
            pass
        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (1, 0, 1)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (1, 0, 1, True)

        # Iteration 2 - close
        it = r1.iter_data()
        next(it)
        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (0, 1, 1)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (0, 1, 1, True)

        it.close()
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (1, 0, 1, True)

        # Iteration 3 - del
        it = r1.iter_data()
        next(it)
        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (0, 1, 1)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (0, 1, 1, True)

        del it
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (1, 0, 1, True)

        # Iteration 4 - try source.close
        it = r1.iter_data()
        next(it)
        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (0, 1, 1)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (0, 1, 1, True)

        with pytest.raises(Exception, match='deactivate'):
            r1.close()
        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (0, 1, 1)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (0, 1, 1, True)

        del it
        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (1, 0, 1)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (1, 0, 1, True)

        # Iteration 5 - multi
        it0 = r1.iter_data()
        next(it0)
        it1 = r1.iter_data()
        next(it1)
        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (0, 2, 2)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (0, 2, 2, True)

        it2 = r1.iter_data()
        with pytest.raises(RuntimeError, match='simultaneous'):
            next(it2)

        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (0, 2, 2)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (0, 2, 2, True)

        r1.activate() # no effect
        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (0, 2, 2)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (0, 2, 2, True)

        del it0
        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (1, 1, 2)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (1, 1, 2, True)

        del it1
        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (2, 0, 2)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (2, 0, 2, True)

def test_vector_concurrent():

    def _work(i):
        point, = r1.iter_data(None)
        return point

    ds = buzz.Dataset(max_active=4)
    meta = dict(
        type='point',
    )

    p = mp.pool.ThreadPool(4)
    with ds.acreate_vector('/tmp/v1.shp', **meta).delete as r1:
        pt = sg.Point([42, 45])
        r1.insert_data(pt)
        r1.deactivate()

        points = list(p.map(_work, range(1000)))
        assert all(p == pt for p in points)
        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (4, 0, 4)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (4, 0, 4, True)

    p.terminate()

def test_raster():
    ds = buzz.DataSource(max_activated=2) # Test deprecated name
    meta = dict(
        fp = buzz.Footprint(
            tl=(1, 1),
            size=(10, 10),
            rsize=(10, 10),
        ),
        dtype=float,
        channel_count=1,
    )

    def statuses(*args):
        l = tuple([
            (
                ds._back.idle_count(prox._back.uid),
                ds._back.used_count(prox._back.uid),
                prox.active_count,
                prox.active
            )
            for prox in args
        ])
        return l

    assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (0, 0, 0)
    with ds.acreate_raster('/tmp/t1.tif', **meta).delete as r1:
        assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (1, 0, 1)
        assert (ds._back.idle_count(r1._back.uid), ds._back.used_count(r1._back.uid), r1.active_count, r1.active) == (1, 0, 1, True)
        with ds.acreate_raster('/tmp/t2.tif', **meta).delete as r2:
            assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (2, 0, 2)
            assert statuses(r1, r2) == ((1, 0, 1, True), (1, 0, 1, True))

            with ds.acreate_raster('/tmp/t3.tif', **meta).delete as r3:
                assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (2, 0, 2)
                assert statuses(r1, r2, r3) == ((0, 0, 0, False), (1, 0, 1, True), (1, 0, 1, True))

                r1.activate()
                assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (2, 0, 2)
                assert statuses(r1, r2, r3) == ((1, 0, 1, True), (0, 0, 0, False), (1, 0, 1, True))

                r1.deactivate()
                r2.activate()
                assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (2, 0, 2)
                assert statuses(r1, r2, r3) == ((0, 0, 0, False), (1, 0, 1, True), (1, 0, 1, True))

                with pytest.raises(RuntimeError, match='max_activated'):
                    ds.activate_all()
                assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (2, 0, 2)
                assert statuses(r1, r2, r3) == ((0, 0, 0, False), (1, 0, 1, True), (1, 0, 1, True))

                ds.deactivate_all()
                assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (0, 0, 0)
                assert statuses(r1, r2, r3) == ((0, 0, 0, False), (0, 0, 0, False), (0, 0, 0, False))

                r1.fill(42)
                r2.fill(42)
                r3.fill(42)
                assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (2, 0, 2)
                assert statuses(r1, r2, r3) == ((0, 0, 0, False), (1, 0, 1, True), (1, 0, 1, True))

            assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (1, 0, 1)
            assert statuses(r1, r2) == ((0, 0, 0, False), (1, 0, 1, True))
            ds.deactivate_all()
            ds.activate_all()
            assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (2, 0, 2)
            assert statuses(r1, r2) == ((1, 0, 1, True), (1, 0, 1, True))

            return

    assert (ds._back.idle_count(), ds._back.used_count(), ds.active_count) == (0, 0, 0)



"""help(Analysis)"""

import numpy as np

from buzzard import Footprint
from buzzard import _tools

class Analysis(object):
    """Private class used to assess the quality of a coordinate transformation"""

    def __init__(self, transformation, inverse, rect=None):
        self.messages = []
        self.conversions_finite = None
        self.inverse_valid = None
        self.angles_valid = None
        self.ratio_valid = None

        if isinstance(rect, Footprint):
            self._run_rect_analysis(transformation, inverse, _tools.Rect(*rect.coords))
        elif np.asarray(rect).shape == (4,):
            minx, maxx, miny, maxy = rect
            rect = _tools.Rect(
                (minx, maxy),
                (minx, miny),
                (maxx, miny),
                (maxx, maxy),
            )
            self._run_rect_analysis(transformation, inverse, rect)

    def _run_rect_analysis(self, transformation, inverse, rect1):
        # *************************************************************************************** **
        self.conversions_finite = False
        rect2 = _tools.Rect(*np.asarray(transformation(rect1.coords))[:, :2])
        if not np.isfinite(rect2.coords).all():
            self.messages.append('forward transformation yielded an infite or nan value')
            return
        rect3 = _tools.Rect(*np.asarray(inverse(rect2.coords))[:, :2])
        if not np.isfinite(rect3.coords).all():
            self.messages.append('inverse transformation yielded an infite or nan value')
            return
        self.conversions_finite = True

        # *************************************************************************************** **
        self.inverse_valid = False
        if (np.abs(rect1.coords - rect3.coords) > rect1.spatial_precision).any():
            self.messages.append(
                'inverse transformation resulted in an error of {} (should be <{})'.format(
                    np.max(np.abs(rect3.coords - rect1.coords)), rect1.spatial_precision,
                )
            )
            return
        self.inverse_valid = True

        # *************************************************************************************** **
        self.angles_valid = False

        if rect1.empty:
            self.messages.append(
                'rectangle is empty (height:{}, width:{})'.format(
                    np.linalg.norm(rect1.tl - rect1.bl),
                    np.linalg.norm(rect1.tl - rect1.tr),
                )
            )
            return

        slack_angles = rect2.tr_slack_angles
        assert slack_angles[0] < slack_angles[1]
        if np.prod(np.sign(slack_angles - 90)) != -1:
            self.messages.append(
                'tl-tr-br angle is between {} and {} degree (should be <90 and >90)'.format(*slack_angles)
                )
            return
        slack_angles = rect2.tl_slack_angles
        assert slack_angles[0] < slack_angles[1]
        if np.prod(np.sign(slack_angles - 90)) != -1:
            self.messages.append(
                'bl-tl-tr angle is between {} and {} degree (should be <90 and >90)'.format(*slack_angles)
                )
            return
        slack_angles = rect2.bl_slack_angles
        assert slack_angles[0] < slack_angles[1]
        if np.prod(np.sign(slack_angles - 90)) != -1:
            self.messages.append(
                'br-bl-tl angle is between {} and {} degree (should be <90 and >90)'.format(*slack_angles)
                )
            return
        slack_angles = rect2.br_slack_angles
        assert slack_angles[0] < slack_angles[1]
        if np.prod(np.sign(slack_angles - 90)) != -1:
            self.messages.append(
                'tr-br-bl angle is between {} and {} degree (should be <90 and >90)'.format(*slack_angles)
                )
            return
        self.angles_valid = True

        # *************************************************************************************** **
        self.ratio_valid = False
        ratiomin, ratiomax = rect1.ratio_slaking
        if not ratiomin < rect2.ratio < ratiomax:
            self.messages.append(
                'rectangle proportion is {} and should be between {} and {}'.format(
                    rect2.ratio, ratiomin, ratiomax)
            )
            return
        self.ratio_valid = True
        # *************************************************************************************** **




# pylint: disable=redefined-outer-name

from __future__ import division, print_function
import itertools

import shapely.geometry as sg
import numpy as np
import pytest

import buzzard as buzz

LESS_ERROR = 1 / 3
MORE_ERROR = 2

FP1 = buzz.Footprint(
    tl=(672939.369686, 6876118.107215),
    size=(24.020000, 24.020000),
    rsize=(1201, 1201),
)

def pytest_generate_tests(metafunc):
    with buzz.Env(significant=15):
        metafunc.parametrize(
            argnames='fp',
            argvalues=[
                FP1,
                FP1.move((-FP1.diagvec / 2)),
                FP1.move((1e8, 1e8)),
                FP1.move((1e8, 1e8), (1e8 + 1, 1e8), (1e8 + 1, 1e8 - 1)),
            ],
        )

@pytest.fixture(params=[6, 8, 10])
def env(request):
    with buzz.Env(significant=request.param):
        yield request.param

def test_same_grid_and_move(fp, env):
    ofp = fp
    if env < ofp._significant_min + 1:
        pytest.skip()

    for factx, facty in itertools.combinations_with_replacement([-1, 0, +1], 2):
        tl = ofp.tl % ofp.pxsize + 1e9 // ofp.pxsize * ofp.pxsize * [factx, facty]
        fp = ofp.move(tl)
        if env < fp._significant_min:
            continue
        eps = np.abs(np.r_[fp.coords, ofp.coords]).max() * 10 ** -buzz.env.significant

        for factx, facty in itertools.combinations_with_replacement([-1, 0, +1], 2):
            fact = np.asarray([factx, facty])

            fp = ofp.move(tl + eps * LESS_ERROR * fact)
            assert ofp.same_grid(fp)

            if (fact != 0).any():
                fp = ofp.move(tl + eps * MORE_ERROR * fact)
                assert not ofp.same_grid(fp)

def test_intersection_and_equals_and_of_extent(fp, env):
    if env < fp._significant_min:
        pytest.skip()
    eps = np.abs(fp.coords).max() * 10 ** -buzz.env.significant
    cwr = itertools.combinations_with_replacement

    for ax, ay, bx, by in cwr([-eps * LESS_ERROR, 0, +eps * LESS_ERROR], 4):
        deltas = np.asarray([ax, ay, bx, by])

        assert fp.almost_equals(fp & sg.LineString([fp.tl + [ax, ay], fp.br + [bx, by]]))
        assert fp.almost_equals(fp.of_extent(fp.extent + deltas, fp.scale))

        if (np.asarray([ax, ay, bx, by]) != 0).any():
            assert fp != fp.of_extent(fp.extent + deltas / LESS_ERROR * MORE_ERROR, fp.scale)

    for slacka, slackb in itertools.product(
            [0,
             -fp.pxvec / np.linalg.norm(fp.pxvec) * eps * MORE_ERROR,
             -fp.pxlrvec / np.linalg.norm(fp.pxlrvec) * eps * MORE_ERROR,
             -fp.pxtbvec / np.linalg.norm(fp.pxtbvec) * eps * MORE_ERROR],
            [0,
             fp.pxvec / np.linalg.norm(fp.pxvec) * eps * MORE_ERROR,
             fp.pxlrvec / np.linalg.norm(fp.pxlrvec) * eps * MORE_ERROR,
             fp.pxtbvec / np.linalg.norm(fp.pxtbvec) * eps * MORE_ERROR],
    ):
        if np.isscalar(slacka) and np.isscalar(slackb):
            continue
        assert fp != fp.dilate(2) & sg.LineString([fp.tl + slacka, fp.br + slackb])


def test_spatial_to_raster(fp, env):
    if env < fp._significant_min:
        pytest.skip()
    rng = np.random.RandomState(42)
    eps = np.abs(fp.coords).max() * 10 ** -buzz.env.significant

    xy = np.dstack(fp.meshgrid_spatial)
    rxy = np.dstack(fp.meshgrid_raster)
    res = np.equal(
        rxy, fp.spatial_to_raster(xy),
    )
    assert np.all(res)
    res = np.equal(
        rxy, fp.spatial_to_raster(xy + (rng.rand(*xy.shape) * 2 - 1) * eps * LESS_ERROR),
    )
    assert np.all(res)
    res = np.equal(
        rxy, fp.spatial_to_raster(xy + (rng.rand(*xy.shape) * 2 - 1) * eps * MORE_ERROR),
    )
    assert not np.all(res)



"""Test tools"""

from __future__ import division, print_function
import itertools
import logging

import numpy as np
from osgeo import gdal, osr

import buzzard as buzz

LOGGER = logging.getLogger('buzzard')

_SRS = [
    ('EPSG:2154', 2154, 489353.59, 6587552.20, -378305.81, 6093283.21, 1212610.74, 7186901.68, """PROJCS["RGF93 / Lambert-93",GEOGCS["RGF93",DATUM["Reseau_Geodesique_Francais_1993",SPHEROID["GRS 1980",6378137,298.257222101,AUTHORITY["EPSG","7019"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY["EPSG","6171"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4171"]],PROJECTION["Lambert_Conformal_Conic_2SP"],PARAMETER["standard_parallel_1",49],PARAMETER["standard_parallel_2",44],PARAMETER["latitude_of_origin",46.5],PARAMETER["central_meridian",3],PARAMETER["false_easting",700000],PARAMETER["false_northing",6600000],UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["X",EAST],AXIS["Y",NORTH],AUTHORITY["EPSG","2154"]]"""), # pylint: disable=line-too-long
    ('EPSG:2154 centered', 0, 0, 0,-867659.4, -494268.99, 723257.15, 599349.48, """PROJCS["RGF93 / Lambert-93",GEOGCS["RGF93",DATUM["Reseau_Geodesique_Francais_1993",SPHEROID["GRS 1980",6378137,298.257222101,AUTHORITY["EPSG","7019"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY["EPSG","6171"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4171"]],PROJECTION["Lambert_Conformal_Conic_2SP"],PARAMETER["standard_parallel_1",49],PARAMETER["standard_parallel_2",44],PARAMETER["latitude_of_origin",46.5],PARAMETER["central_meridian",3],PARAMETER["false_easting",210646.41],PARAMETER["false_northing",12447.79999999981],UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["X",EAST],AXIS["Y",NORTH]]"""), # pylint: disable=line-too-long
    ('EPSG:27561', 27561, 552587.60, 215784.19, 64398.20, 74408.53, 1012167.98, 398560.88, """PROJCS["NTF (Paris) / Lambert Nord France",GEOGCS["NTF (Paris)",DATUM["Nouvelle_Triangulation_Francaise_Paris",SPHEROID["Clarke 1880 (IGN)",6378249.2,293.4660212936265,AUTHORITY["EPSG","7011"]],TOWGS84[-168,-60,320,0,0,0,0],AUTHORITY["EPSG","6807"]],PRIMEM["Paris",2.33722917,AUTHORITY["EPSG","8903"]],UNIT["grad",0.01570796326794897,AUTHORITY["EPSG","9105"]],AUTHORITY["EPSG","4807"]],PROJECTION["Lambert_Conformal_Conic_1SP"],PARAMETER["latitude_of_origin",55],PARAMETER["central_meridian",0],PARAMETER["scale_factor",0.999877341],PARAMETER["false_easting",600000],PARAMETER["false_northing",200000],UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["X",EAST],AXIS["Y",NORTH],AUTHORITY["EPSG","27561"]]"""), # pylint: disable=line-too-long
    ('EPSG:5627', 5627, 836179.68, 5889250.81, -927864.00, 3972451.32, 1395572.16, 8094223.77, """PROJCS["ED50 / TM 6 NE",GEOGCS["ED50",DATUM["European_Datum_1950",SPHEROID["International 1924",6378388,297,AUTHORITY["EPSG","7022"]],TOWGS84[-87,-98,-121,0,0,0,0],AUTHORITY["EPSG","6230"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4230"]],PROJECTION["Transverse_Mercator"],PARAMETER["latitude_of_origin",0],PARAMETER["central_meridian",6],PARAMETER["scale_factor",0.9996],PARAMETER["false_easting",500000],PARAMETER["false_northing",0],UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["Easting",EAST],AXIS["Northing",NORTH],AUTHORITY["EPSG","5627"]]"""), # pylint: disable=line-too-long
    ('EPSG:3950', 3950, 1488760.97, 8798334.99, 613684.96, 8306333.45, 2211114.59, 9398784.57, """PROJCS["RGF93 / CC50",GEOGCS["RGF93",DATUM["Reseau_Geodesique_Francais_1993",SPHEROID["GRS 1980",6378137,298.257222101,AUTHORITY["EPSG","7019"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY["EPSG","6171"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4171"]],PROJECTION["Lambert_Conformal_Conic_2SP"],PARAMETER["standard_parallel_1",49.25],PARAMETER["standard_parallel_2",50.75],PARAMETER["latitude_of_origin",50],PARAMETER["central_meridian",3],PARAMETER["false_easting",1700000],PARAMETER["false_northing",9200000],UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["X",EAST],AXIS["Y",NORTH],AUTHORITY["EPSG","3950"]]"""), # pylint: disable=line-too-long
]

SRS = [
    dict(
        name=name,
        epsg=epsg,
        cx=cx, cy=cy, minx=minx, miny=miny, maxx=maxx, maxy=maxy, wkt=wkt,
        w=maxx - minx, h=maxy - miny
    )
    for name, epsg, cx, cy, minx, miny, maxx, maxy, wkt in _SRS
]


def get_srs_by_name(name):
    return next(
        srs
        for srs in SRS
        if srs['name'] == name
    )

def eq(*items, **kwargs):
    """Numeric items are all almost equal"""
    tol = kwargs.pop('tol', 10e-5)
    assert not kwargs
    for a, b in itertools.combinations(items, 2):
        diff = np.abs(np.asarray(a) - np.asarray(b))
        if not (diff <= tol).all():
            return False
    return True


def eqall(items, **kwargs):
    """Numeric items are all almost equal"""
    return eq(*items, **kwargs)


def fpeq(*items, **kwargs):
    """Footprint items are all almost equal"""
    tol = kwargs.pop('tol', 10e-5)
    assert not kwargs
    for a, b in itertools.combinations(items, 2):
        diff = np.abs(a.gt - b.gt)
        if not (diff <= tol).all():
            return False
        diff = np.abs(a.rsize - b.rsize)
        if not (diff <= tol).all():
            return False
    return True

def sreq(*items, verbose_on=None):
    """SR items are all equal"""
    v = tuple(map(int, gdal.__version__.split('.')))
    for a, b in itertools.combinations(items, 2):
        a = osr.SpatialReference(osr.GetUserInputAsWKT(a))
        if v < (3,):
            a.StripCTParms()
        a = a.ExportToProj4()
        a = osr.SpatialReference(osr.GetUserInputAsWKT(a))
        if v < (3,):
            a.StripCTParms()

        b = osr.SpatialReference(osr.GetUserInputAsWKT(b))
        if v < (3,):
            b.StripCTParms()
        b = b.ExportToProj4()
        b = osr.SpatialReference(osr.GetUserInputAsWKT(b))
        if v < (3,):
            b.StripCTParms()

        res = bool(a.IsSame(b))
        if res is verbose_on:
            print('')
            print(a.ExportToPrettyWkt())
            print('---------- vs ----------')
            print(b.ExportToPrettyWkt())
            print('')

        if not res:
            return False
    return True

def poly_relation(a, b):
    """Describe 2 polygons relation"""
    rels = {
        'a contains b': a.contains(b),
        'a crosses b': a.crosses(b),
        'a disjoint b': a.disjoint(b),
        'a intersects b': a.intersects(b),
        'a touches b': a.touches(b),
        'a within b': a.within(b),
        'a covers b': a.covers(b),
        'a overlaps b': a.overlaps(b),
        'b contains a': b.contains(a),
        'b crosses a': b.crosses(a),
        'b disjoint a': b.disjoint(a),
        'b intersects a': b.intersects(a),
        'b touches a': b.touches(a),
        'b within a': b.within(a),
        'b covers a': b.covers(a),
        'b overlaps a': b.overlaps(a),
    }
    return ', '.join(
        k for k, v in rels.items() if v
    )

ROOT_TL = np.asarray([296455., 71495.])


def make_tif(path, tloffset=(0, 0), reso=(0.25, -0.25), rsize=(20, 10),
             proj=SRS[0]['wkt'], channel_count=1, dtype=gdal.GDT_Float32):
    """Create a tiff files and return info about it"""
    tl = ROOT_TL + tloffset
    reso = np.asarray(reso)
    fp = buzz.Footprint(tl=tl, rsize=rsize, size=np.abs(reso * rsize))
    x, y = fp.meshgrid_spatial
    x = np.abs(x) - abs(ROOT_TL[0])
    y = abs(ROOT_TL[1]) - np.abs(y)
    x *= 15
    y *= 15
    a = x / 2 + y / 2
    a = np.around(a).astype('float32')
    driver = gdal.GetDriverByName('GTiff')
    dataset = driver.Create(path, rsize[0], rsize[1], channel_count, dtype)
    dataset.SetGeoTransform(fp.gt)
    dataset.SetProjection(proj)
    for i in range(channel_count):
        dataset.GetRasterBand(i + 1).WriteArray(a)
        dataset.GetRasterBand(i + 1).SetNoDataValue(-32000.)
    dataset.FlushCache()
    return path, fp, a


def make_tif2(path, reso=(1., -1.), rsize=(10, 10), tl=(0., 10.),
              proj=SRS[0]['wkt'], channel_count=1, dtype=gdal.GDT_Float32,
              nodata=-32000, nodata_border_size=(0, 0, 0, 0)):
    """Create a tiff files"""
    reso = np.asarray(reso)
    fp = buzz.Footprint(tl=tl, rsize=rsize, size=np.abs(reso * rsize))
    x, y = fp.meshgrid_raster
    a = x + y
    if nodata_border_size != 0:
        l, r, t, b = nodata_border_size
        if t != 0:
            a[None:t, None:None] = nodata
        if b != 0:
            a[-b:None, None:None] = nodata
        if l != 0:
            a[None:None, None:l] = nodata
        if r != 0:
            a[None:None, -r:None] = nodata

    LOGGER.info('TIFF ARRAY:%s\n', a)
    gdal.UseExceptions()
    driver = gdal.GetDriverByName('GTiff')
    dataset = driver.Create(path, int(rsize[0]), int(rsize[1]), channel_count, dtype)
    dataset.SetGeoTransform(fp.gt)
    dataset.SetProjection(proj)
    for i in range(channel_count):
        dataset.GetRasterBand(i + 1).WriteArray(a)
        dataset.GetRasterBand(i + 1).SetNoDataValue(nodata)
    dataset.FlushCache()

def dump_tiles(tiles):
    """Print tiles"""
    for y, x in itertools.product(range(tiles.shape[0]), range(tiles.shape[1])):
        print(tiles[y, x])

def assert_tiles_eq(mata, matb, tol=10e-5):
    """Assert that two matrices contain the same tiles"""
    mata = np.asarray(mata)
    matb = np.asarray(matb)

    def _dump():
        print('--------')
        dump_tiles(mata)
        print('--------')
        dump_tiles(matb)

    if not mata.shape == matb.shape:
        _dump()
        assert mata.shape == matb.shape
    for a, b in zip(mata.flatten(), matb.flatten()):
        if not fpeq(a, b, tol=tol):
            _dump()
            assert fpeq(a, b, tol=tol)



import collections
import multiprocessing as mp
import multiprocessing.pool
import functools

import numpy as np

from buzzard._actors.message import Msg
from buzzard._actors.pool_job import ProductionJobWaiting, PoolJobWorking

class ActorComputer(object):
    """Actor that takes care of sheduling computations by using user's `compute_array` function"""

    def __init__(self, raster):
        self._raster = raster
        self._alive = True
        computation_pool = raster.computation_pool
        if computation_pool is not None:
            self._waiting_room_address = '/Pool{}/WaitingRoom'.format(id(computation_pool))
            self._working_room_address = '/Pool{}/WorkingRoom'.format(id(computation_pool))
            if isinstance(computation_pool, mp.pool.ThreadPool):
                self._same_address_space = True
            elif isinstance(computation_pool, mp.pool.Pool):
                self._same_address_space = False
            else: # pragma: no cover
                assert False, 'Type should be checked in facade'
        self._waiting_jobs_per_query = collections.defaultdict(set)
        self._working_jobs = set()

        self._performed_computations = set() # type: Set[Footprint]
        self.address = '/Raster{}/Computer'.format(self._raster.uid)

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_compute_this_array(self, qi, compute_idx):
        """Receive message: Start making this array"""
        msgs = []

        if self._raster.computation_pool is None:
            work = self._create_work_job(qi, compute_idx)
            compute_fp = qi.cache_computation.list_of_compute_fp[compute_idx]
            if compute_fp not in self._performed_computations:
                res = work.func()
                res = self._normalize_user_result(compute_fp, res)
                self._raster.debug_mngr.event('object_allocated', res)
                self._performed_computations.add(compute_fp)
                msgs += self._commit_work_result(work, res)

        else:
            wait = Wait(self, qi, compute_idx)
            self._waiting_jobs_per_query[qi].add(wait)
            msgs += [Msg(self._waiting_room_address, 'schedule_job', wait)]

        return msgs

    def receive_token_to_working_room(self, job, token):
        msgs = []

        self._waiting_jobs_per_query[job.qi].remove(job)
        if len(self._waiting_jobs_per_query[job.qi]) == 0:
            del self._waiting_jobs_per_query[job.qi]

        work = self._create_work_job(job.qi, job.compute_idx)

        compute_fp = job.qi.cache_computation.list_of_compute_fp[job.compute_idx]
        if compute_fp not in self._performed_computations:
            msgs += [Msg(self._working_room_address, 'launch_job_with_token', work, token)]
            self._performed_computations.add(compute_fp)
            self._working_jobs.add(work)
        else:
            msgs += [Msg(self._working_room_address, 'salvage_token', token)]

        return msgs

    def receive_job_done(self, job, result):
        result = self._normalize_user_result(job.compute_fp, result)
        self._raster.debug_mngr.event('object_allocated', result)
        self._working_jobs.remove(job)
        return self._commit_work_result(job, result)

    def receive_cancel_this_query(self, qi):
        """Receive message: One query was dropped

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        """
        msgs = []
        for job in self._waiting_jobs_per_query[qi]:
            msgs += [Msg(self._waiting_room_address, 'unschedule_job', job)]
        del self._waiting_jobs_per_query[qi]
        return msgs

    def receive_die(self):
        """Receive message: The raster was killed"""
        assert self._alive
        self._alive = False

        msgs = []
        msgs += [
            Msg(self._waiting_room_address, 'unschedule_job', job)
            for jobs in self._waiting_jobs_per_query.values()
            for job in jobs
        ]
        self._waiting_jobs_per_query.clear()

        msgs += [
            Msg(self._working_room_address, 'cancel_job', job)
            for job in self._working_jobs
        ]
        self._working_jobs.clear()

        self._raster = None
        return msgs

    # ******************************************************************************************* **
    def _create_work_job(self, qi, compute_idx):
        return Work(
            self, qi, compute_idx,
        )

    def _commit_work_result(self, work_job, res):
        return [Msg('ComputationAccumulator', 'combine_this_array', work_job.compute_fp, res)]

    def _normalize_user_result(self, compute_fp, res):
        if not isinstance(res, np.ndarray): # pragma: no cover
            raise ValueError("Result of recipe's `compute_array` have type {}, it should be ndarray".format(
                type(res)
            ))
        res = np.atleast_3d(res)
        y, x, c = res.shape
        if (y, x) != tuple(compute_fp.shape): # pragma: no cover
            raise ValueError("Result of recipe's `compute_array` have shape `{}`, should start with {}".format(
                res.shape,
                compute_fp.shape,
            ))
        if c != len(self._raster): # pragma: no cover
            raise ValueError("Result of recipe's `compute_array` have shape `{}`, should have {} bands".format(
                res.shape,
                len(self._raster),
            ))
        res = res.astype(self._raster.dtype, copy=False)
        return res

    # ******************************************************************************************* **

class Wait(ProductionJobWaiting):

    def __init__(self, actor, qi, compute_idx):
        self.qi = qi
        self.compute_idx = compute_idx
        qicc = qi.cache_computation

        compute_fp = qicc.list_of_compute_fp[compute_idx]
        prod_idx = qicc.dict_of_min_prod_idx_per_compute_fp[compute_fp]
        super().__init__(actor.address, qi, prod_idx, 4, compute_fp)

class Work(PoolJobWorking):
    def __init__(self, actor, qi, compute_idx):
        qicc = qi.cache_computation
        assert qicc.collected_count == compute_idx, (qicc.collected_count, compute_idx)

        compute_fp = qicc.list_of_compute_fp[compute_idx]

        self.compute_fp = compute_fp

        primitive_arrays = {}
        primitive_footprints = {}
        for prim_name, queue in qicc.primitive_queue_per_primitive.items():
            primitive_arrays[prim_name] = queue.get_nowait()
            primitive_footprints[prim_name] = qicc.primitive_fps_per_primitive[prim_name][compute_idx]

        qicc.collected_count += 1

        if actor._raster.computation_pool is None or actor._same_address_space:
            func = functools.partial(
                actor._raster.compute_array,
                compute_fp,
                primitive_footprints,
                primitive_arrays,
                actor._raster.facade_proxy
            )
        else:
            func = functools.partial(
                actor._raster.compute_array,
                compute_fp,
                primitive_footprints,
                primitive_arrays,
                None,
            )
        actor._raster.debug_mngr.event('object_allocated', func)

        super().__init__(actor.address, func)



import contextlib

from buzzard._a_emissary_raster import AEmissaryRaster, ABackEmissaryRaster
from buzzard._a_gdal_raster import ABackGDALRaster
from buzzard._tools import conv
from buzzard._footprint import Footprint

class GDALMemRaster(AEmissaryRaster):
    """Concrete class defining the behavior of a GDAL raster using the "MEM" driver.

    >>> help(Dataset.create_raster)

    Features Defined
    ----------------
    None
    """

    def __init__(self, ds, fp, dtype, channel_count, channels_schema, open_options, sr):
        back = BackGDALMemRaster(
            ds._back, fp, dtype, channel_count, channels_schema, open_options, sr,
        )
        super(GDALMemRaster, self).__init__(ds=ds, back=back)

class BackGDALMemRaster(ABackEmissaryRaster, ABackGDALRaster):
    """Implementation of GDALMemRaster"""

    def __init__(self, back_ds, fp, dtype, channel_count, channels_schema, open_options, sr):

        gdal_ds = self.create_file(
            '', fp, dtype, channel_count, channels_schema, 'MEM', open_options, sr, False
        )
        self._gdal_ds = gdal_ds

        path = gdal_ds.GetDescription()
        driver = gdal_ds.GetDriver().ShortName
        fp_stored = Footprint(
            gt=gdal_ds.GetGeoTransform(),
            rsize=(gdal_ds.RasterXSize, gdal_ds.RasterYSize),
        )
        channels_schema = self._channels_schema_of_gdal_ds(gdal_ds)
        dtype = conv.dtype_of_gdt_downcast(gdal_ds.GetRasterBand(1).DataType)
        sr = gdal_ds.GetProjection()
        if sr == '':
            wkt_stored = None
        else:
            wkt_stored = sr

        super(BackGDALMemRaster, self).__init__(
            back_ds=back_ds,
            wkt_stored=wkt_stored,
            channels_schema=channels_schema,
            dtype=dtype,
            fp_stored=fp_stored,
            mode='w',
            driver=driver,
            open_options=open_options,
            path=path,
        )

    @contextlib.contextmanager
    def acquire_driver_object(self):
        yield self._gdal_ds

    def delete(self): # pragma: no cover
        raise NotImplementedError('GDAL MEM driver does no allow deletion, use `close`')

    def close(self):
        super(BackGDALMemRaster, self).close()
        del self._gdal_ds



""">>> help(TileMixin)"""

import numpy as np
from buzzard._env import env

class MoveMixin(object):
    """Private mixin for the Footprint class containing move subroutines"""

    def _snap_target_coordinates_before_move(self, tl1, tr1, br1):
        rw, rh = self.rsize

        # Extract `source` values *********************************************** **
        v0 = self.pxlrvec
        w0 = self.pxtbvec
        norm_v0, norm_w0 = self.pxsize
        i0 = v0 / norm_v0
        j0 = w0 / norm_w0

        # Compute `transformed values` ****************************************** **
        v1 = (tr1 - tl1) / rw
        w1 = (br1 - tr1) / rh
        norm_v1 = np.linalg.norm(v1)
        norm_w1 = np.linalg.norm(w1)
        i1 = v1 / norm_v1
        j1 = w1 / norm_w1

        # Prepare comparison function ******************************************* **
        largest_coord = np.abs([tl1, tr1, br1]).max()
        spatial_precision = largest_coord * 10 ** -env.significant
        def _are_close_enough(p, q):
            return (np.abs(p - q) < spatial_precision).all()

        # Compute `transformed and snapped values` ****************************** **
        tl2 = tl1

        # Attempt rounding to preserve `angle` and `pxsize`
        # e.g. The transformation is just a shift
        # e.g. The transformation is a flip or a mirror
        i2 = np.copysign(i0, i1)
        j2 = np.copysign(j0, j1)
        norm_v2 = norm_v0
        norm_w2 = norm_w0

        v2 = i2 * norm_v2
        w2 = j2 * norm_w2
        tr2 = tl2 + v2 * rw
        br2 = tr2 + w2 * rh
        if _are_close_enough(tr1, tr2) and _are_close_enough(br1, br2):
            return tl2, tr2, br2

        # Attempt rounding to preserve `angle` and `pxsizex / pxsizey`
        # e.g. The transformation is a change of unit
        i2 = np.copysign(i0, i1)
        j2 = np.copysign(j0, j1)
        norm_v2 = norm_v1
        norm_w2 = norm_v1 / norm_v0 * norm_w0

        v2 = i2 * norm_v2
        w2 = j2 * norm_w2
        tr2 = tl2 + v2 * rw
        br2 = tr2 + w2 * rh
        if _are_close_enough(tr1, tr2) and _are_close_enough(br1, br2):
            return tl2, tr2, br2

        # Attempt rounding to preserve `angle`
        # e.g. The transformation is a change of unit different on each axis
        i2 = np.copysign(i0, i1)
        j2 = np.copysign(j0, j1)
        norm_v2 = norm_v1
        norm_w2 = norm_w1

        v2 = i2 * norm_v2
        w2 = j2 * norm_w2
        tr2 = tl2 + v2 * rw
        br2 = tr2 + w2 * rh
        if _are_close_enough(tr1, tr2) and _are_close_enough(br1, br2):
            return tl2, tr2, br2

        # Attempt rounding to preserve `pxsize`
        # e.g. The transformation is a rotation
        i2 = i1
        j2 = j1
        norm_v2 = norm_v0
        norm_w2 = norm_w0

        v2 = i2 * norm_v2
        w2 = j2 * norm_w2
        tr2 = tl2 + v2 * rw
        br2 = tr2 + w2 * rh
        if _are_close_enough(tr1, tr2) and _are_close_enough(br1, br2):
            return tl2, tr2, br2

        # Attempt rounding to preserve `pxsizex / pxsizey`
        # e.g. The transformation is a rotation and a change of unit
        i2 = i1
        j2 = j1
        norm_v2 = norm_v1
        norm_w2 = norm_v1 / norm_v0 * norm_w0

        v2 = i2 * norm_v2
        w2 = j2 * norm_w2
        tr2 = tl2 + v2 * rw
        br2 = tr2 + w2 * rh
        if _are_close_enough(tr1, tr2) and _are_close_enough(br1, br2):
            return tl2, tr2, br2

        # Don't perform snapping
        return tl1, tr1, br1



from buzzard._actors.message import Msg

import collections

class ActorProducer(object):
    """Actor that takes care of waiting for cache tiles reads and launching resamplings"""

    def __init__(self, raster):
        self._raster = raster
        self._alive = True

        self._produce_per_query = collections.defaultdict(dict) # type: Mapping[CachedQueryInfos, Mapping[int, _ProdArray]]
        self.address = '/Raster{}/Producer'.format(self._raster.uid)

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_make_this_array(self, qi, prod_idx):
        """Receive message: Start making this array

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        prod_idx: int
        """
        msgs = []

        pi = qi.prod[prod_idx] # type: CacheProduceInfos
        assert pi.share_area is (len(pi.cache_fps) != 0)

        if pi.share_area:
            # If this prod_idx requires some cache file reads (this is the case most of the time)
            msgs += [Msg(
                'CacheExtractor', 'sample_those_cache_files_to_an_array', qi, prod_idx,
            )]

        for resample_fp in pi.resample_fps:
            sample_fp = pi.resample_sample_dep_fp[resample_fp]
            if sample_fp is None:
                # Start the 'resampling' step of the resample_fp fully outside of raster
                assert (
                    resample_fp not in pi.resample_cache_deps_fps or
                    len(pi.resample_cache_deps_fps[resample_fp]) == 0
                )
                msgs += [Msg(
                    'Resampler', 'resample_and_accumulate',
                    qi, prod_idx, None, resample_fp, None,
                )]

        self._produce_per_query[qi][prod_idx] = _ProdArray(pi)
        return msgs

    def receive_sampled_a_cache_file_to_the_array(self, qi, prod_idx, cache_fp, array):
        """Receive message: A cache file was read for that output array

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        prod_idx: int
        cache_fp: Footprint
            The cache_fp that was just read by the reader
        array: ndarray
            The array onto which the reader fills rectangles one by one
        """
        msgs = []

        pr = self._produce_per_query[qi][prod_idx]
        pi = qi.prod[prod_idx]

        # The constraints on `cache_fp` are now satisfied
        for resample_fp, cache_fps in pr.resample_needs.items():
            if cache_fp in cache_fps:
                cache_fps.remove(cache_fp)

        resample_ready = [
            resample_fp
            for resample_fp, cache_fps in pr.resample_needs.items()
            if len(cache_fps) == 0
        ]
        for resample_fp in resample_ready:
            del pr.resample_needs[resample_fp]
            subsample_fp = pi.resample_sample_dep_fp[resample_fp]
            assert subsample_fp is not None
            subsample_array = array[subsample_fp.slice_in(pi.sample_fp)]

            assert subsample_array.shape[:2] == tuple(subsample_fp.shape)
            msgs += [Msg(
                'Resampler', 'resample_and_accumulate',
                qi, prod_idx, subsample_fp, resample_fp, subsample_array,
            )]

        return msgs

    def receive_made_this_array(self, qi, prod_idx, array):
        """Receive message: Done creating an output array"""
        del self._produce_per_query[qi][prod_idx]
        if len(self._produce_per_query[qi]) == 0:
            del self._produce_per_query[qi]
        return [Msg(
            'QueriesHandler', 'made_this_array', qi, prod_idx, array
        )]

    def receive_cancel_this_query(self, qi):
        """Receive message: One query was dropped

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        """
        if qi in self._produce_per_query:
            del self._produce_per_query[qi]
        return []

    def receive_die(self):
        """Receive message: The raster was killed"""
        assert self._alive
        self._alive = False

        self._produce_per_query.clear()
        self._raster = None
        return []

    # ******************************************************************************************* **

class _ProdArray(object):
    def __init__(self, pi):
        self.resample_needs = {
            resample_fp: set(cache_fps)
            for resample_fp, cache_fps in pi.resample_cache_deps_fps.items()
        }



"""Test:
- Raster opening
- Raster creation
- Attributes validity
- Raster deletion
"""

# pylint: disable=redefined-outer-name

from __future__ import division, print_function
import itertools
import os
import uuid
import tempfile

import numpy as np
import pytest
from osgeo import gdal

from .tools import  get_srs_by_name
from buzzard import Footprint, Dataset

SR1 = get_srs_by_name('EPSG:2154')

DRIVER_STORES_SRS = {
    'GTiff': True,
    'BMP': True,
    'BT': True,
    'RMF': False,
    'ERS': False,
    'GTX': False,
    'RST': False,
    'MEM': True,
}

TESTS = [

    dict( # Classic dsm
        subtests=[
            dict(driver='GTiff', ext='.tif', options=[]),
            dict(driver='MEM', options=[]),
            dict(driver='numpy'),
        ],
        dtype='float32',
        channel_count=1,
        channels_schema= {'nodata': [-32767]},
        sr=SR1['wkt'],
        fp=Footprint(tl=(0, 10), size=(10, 10), rsize=(30, 30)),
    ),

    dict( # Classic rgb
        subtests=[
            dict(driver='GTiff', ext='.tif', options=[]),
            dict(driver='MEM', options=[]),
            dict(driver='numpy'),
        ],
        dtype='uint8',
        channel_count=3,
        channels_schema= {},
        sr=SR1['wkt'],
        fp=Footprint(tl=(0, 10), size=(10, 10), rsize=(30, 30)),
    ),

    dict( # Classic rgba
        subtests=[
            dict(driver='GTiff', ext='.tif', options=[]),
            dict(driver='MEM', options=[]),
            dict(driver='numpy'),
        ],
        dtype='uint8',
        channel_count=4,
        channels_schema= {},
        sr=SR1['wkt'],
        fp=Footprint(tl=(0, 10), size=(10, 10), rsize=(30, 30)),
    ),

    dict( # Ununsual channels_schema
        subtests=[
            dict(driver='GTiff', ext='.tif', options=[]),
            dict(driver='MEM', options=[]),
        ],
        dtype='float32',
        channel_count=1,
        channels_schema= {
            'nodata': [42.0],
            'interpretation': ['greenband'],
            'offset': [0.0],
            'scale': [1.0],
            'mask': ['nodata']
        },
        sr=SR1['wkt'],
        fp=Footprint(tl=(0, 10), size=(10, 10), rsize=(30, 30)),
    ),

    dict( # Ununsual channels_schema
        subtests=[
            dict(driver='ERS', ext='.ers', options=[]),
            dict(driver='MEM', options=[]),
        ],
        dtype='float32',
        channel_count=1,
        channels_schema= {
            'nodata': [0.0],
            'interpretation': ['undefined'],
            'offset': [0.0],
            'scale': [1.0],
            'mask': ['nodata']
        },
        sr=SR1['wkt'],
        fp=Footprint(tl=(0, 10), size=(10, 10), rsize=(30, 30)),
    ),

    dict( # Ununsual channels_schema
        subtests=[
            dict(driver='BT', ext='.bt', options=[]),
            dict(driver='GTX', ext='.gtx', options=[]),
            dict(driver='MEM', options=[]),
        ],
        dtype='float32',
        channel_count=1,
        channels_schema= {
            'nodata': [-32768.0],
            'interpretation': ['undefined'],
            'offset': [0.0],
            'scale': [1.0],
            'mask': ['nodata']
        },
        sr=SR1['wkt'],
        fp=Footprint(tl=(0, 10), size=(10, 10), rsize=(30, 30)),
    ),

    dict( # Ununsual channels_schema
        subtests=[
            dict(driver='RST', ext='.rst', options=[]),
            dict(driver='MEM', options=[]),
        ],
        dtype='float32',
        channel_count=1,
        channels_schema= {
            'nodata': [-32768.0],
            'interpretation': ['grayindex'],
            'offset': [0.0],
            'scale': [1.0],
            'mask': ['nodata']
        },
        sr=SR1['wkt'],
        fp=Footprint(tl=(0, 10), size=(10, 10), rsize=(30, 30)),
    ),

    dict( # Ununsual channels_schema
        subtests=[
            dict(driver='GTiff', ext='.tif', options=[]),
            dict(driver='MEM', options=[]),
        ],
        dtype='float64',
        channel_count=1,
        channels_schema= {
            'nodata': [42.0],
            'interpretation': ['greenband'],
            'offset': [0.0],
            'scale': [1.0],
            'mask': ['nodata']
        },
        sr=SR1['wkt'],
        fp=Footprint(tl=(0, 10), size=(10, 10), rsize=(30, 30)),
    ),

    dict( # Ununsual channels_schema
        subtests=[
            dict(driver='RMF', ext='.mtw', options=['MTW=ON']),
            dict(driver='ERS', ext='.ers', options=[]),
            dict(driver='MEM', options=[]),
        ],
        dtype='float64',
        channel_count=1,
        channels_schema= {
            'nodata': [0.0],
            'interpretation': ['undefined'],
            'offset': [0.0],
            'scale': [1.0],
            'mask': ['nodata']
        },
        sr=SR1['wkt'],
        fp=Footprint(tl=(0, 10), size=(10, 10), rsize=(30, 30)),
    ),

    dict( # Ununsual channels_schema
        subtests=[
            dict(driver='GTiff', ext='.tif', options=[]),
            dict(driver='BMP', ext='.bmp', options=[]),
            dict(driver='MEM', options=[]),
        ],
        dtype='uint8',
        channel_count=3,
        channels_schema= {
            'nodata': [None, None, None],
            'interpretation': ['redband', 'greenband', 'blueband'],
            'offset': [0.0, 0.0, 0.0],
            'scale': [1.0, 1.0, 1.0],
            'mask': ['all_valid', 'all_valid', 'all_valid'],
        },
        sr=SR1['wkt'],
        fp=Footprint(tl=(0, 10), size=(10, 10), rsize=(30, 30)),
    ),

    dict( # Ununsual channels_schema
        subtests=[
            dict(driver='BMP', ext='.bmp', options=[]),
            dict(driver='RMF', ext='.rsw', options=['MTW=OFF']),
            dict(driver='MEM', options=[]),
        ],
        dtype='uint8',
        channel_count=3,
        channels_schema= {
            'nodata': [0., 0., 0.],
            'interpretation': ['redband', 'greenband', 'blueband'],
            'offset': [0.0, 0.0, 0.0],
            'scale': [1.0, 1.0, 1.0],
            'mask': ['nodata', 'nodata', 'nodata'],
        },
        sr=SR1['wkt'],
        fp=Footprint(tl=(0, 10), size=(10, 10), rsize=(30, 30)),
    ),
]

def pytest_generate_tests(metafunc):
    tests = []
    for test in TESTS:
        for subtest in test['subtests']:
            subtest = dict(subtest)
            meta = dict(test)
            del meta['subtests']

            if 'meta_file' in metafunc.fixturenames and subtest['driver'] not in {'MEM', 'numpy'}:
                ext = subtest.pop('ext')
                meta.update(subtest)
                tests.append((meta, ext))
            if 'meta_mem' in metafunc.fixturenames and subtest['driver'] == 'MEM':
                meta['path'] = ''
                meta.update(subtest)
                tests.append(meta)
            if 'meta_numpy' in metafunc.fixturenames and subtest['driver'] == 'numpy':
                meta = dict(
                    fp=meta['fp'],
                    array=np.empty(
                        np.r_[meta['fp'].shape, meta['channel_count']],
                        meta['dtype']
                    ),
                    channels_schema=meta['channels_schema'],
                    sr=meta['sr'],
                )
                tests.append(meta)

    if 'meta_file' in metafunc.fixturenames:
        metafunc.parametrize(
            argnames='meta_file,ext',
            argvalues=tests,
        )
    if 'meta_mem' in metafunc.fixturenames:
        metafunc.parametrize(
            argnames='meta_mem',
            argvalues=tests,
        )
    if 'meta_numpy' in metafunc.fixturenames:
        metafunc.parametrize(
            argnames='meta_numpy',
            argvalues=tests,
        )

@pytest.fixture
def path(meta_file, ext):
    path = '{}/{}{}'.format(tempfile.gettempdir(), uuid.uuid4(), ext)
    yield path
    if os.path.isfile(path):
        try:
            gdal.GetDriverByName(meta['driver']).Delete(path)
        except:
            os.remove(path)

def test_file(meta_file, path):
    meta = meta_file
    fp = meta['fp']

    ds = Dataset()
    arr = np.add(*fp.meshgrid_raster)
    arr = np.repeat(arr[..., np.newaxis], meta['channel_count'], -1)

    with ds.acreate_raster(path, **meta).close as r:
        r.set_data(arr)

        if DRIVER_STORES_SRS[meta['driver']]:
            assert r.wkt_stored == meta['sr']
            assert r.wkt_virtual == meta['sr']
        for k, v in meta['channels_schema'].items():
            assert r.channels_schema[k] == v
        assert r.dtype == np.dtype(meta['dtype'])
        assert r.fp_stored == fp
        if 'nodata' in meta['channels_schema']:
            assert r.nodata == meta['channels_schema']['nodata'][0]
        else:
            assert r.nodata == None
        assert len(r) == meta['channel_count']
        assert r.fp == fp
        assert r.mode == 'w'
        assert r.driver == meta['driver']
        assert r.open_options == meta['options']
        assert r.path == path
        assert np.all(r.get_data(channels=slice(None)) == arr)

    assert os.path.isfile(path)
    with ds.aopen_raster(path, driver=meta['driver']).close as r:

        if DRIVER_STORES_SRS[meta['driver']]:
            assert r.wkt_stored == meta['sr']
            assert r.wkt_virtual == meta['sr']
        for k, v in meta['channels_schema'].items():
            assert r.channels_schema[k] == v
        assert r.dtype == np.dtype(meta['dtype'])
        assert r.fp_stored.almost_equals(fp)
        if 'nodata' in meta['channels_schema']:
            assert r.nodata == meta['channels_schema']['nodata'][0]
        else:
            assert r.nodata == None
        assert len(r) == meta['channel_count']
        assert r.fp.almost_equals(fp)
        assert r.mode == 'r'
        assert r.driver == meta['driver']
        assert r.path == path
        assert np.all(r.get_data(channels=slice(None)) == arr)

        with pytest.raises(RuntimeError):
            r.delete()

    # Test error - driver name
    meta = meta_file.copy()
    meta['driver'] = 'Truc42'
    with pytest.raises(ValueError, match='driver'):
        ds.acreate_raster(path, **meta)

    # # Test error - need overwrite
    meta = meta_file.copy()
    meta['ow'] = False
    with pytest.raises(RuntimeError, match='overwrite'):
        ds.acreate_raster(path, **meta)

    # Test deletion
    with ds.aopen_raster(path, driver=meta['driver'], mode='w').delete as r:
        assert r.mode == 'w'
    assert not os.path.exists(path)

    # Test error - open
    with pytest.raises(RuntimeError, match='Could not open'):
        ds.aopen_raster(path, driver=meta['driver'])

    # Test error - bad sr
    meta = meta_file.copy()
    meta['sr'] = "ca n'existe pas"
    with pytest.raises(ValueError, match='wkt'):
        ds.acreate_raster(path, **meta)

    # Test error - failed file deletion for overwrite
    if meta['driver'].lower() == 'gtiff':
        try:
            os.mkdir(path)
            os.chmod(path, int('000', 8))
            meta = meta_file.copy()
            meta['ow'] = True
            with pytest.raises(RuntimeError, match='Could not delete'):
                ds.acreate_raster(path, **meta)
        finally:
            os.chmod(path, int('777', 8))
            os.rmdir(path)

def test_mem(meta_mem):
    meta = meta_mem
    fp = meta['fp']

    ds = Dataset()
    arr = np.add(*fp.meshgrid_raster)
    arr = np.repeat(arr[..., np.newaxis], meta['channel_count'], -1)

    with ds.acreate_raster(**meta).close as r:
        r.set_data(arr)

        assert r.wkt_stored == meta['sr']
        assert r.wkt_virtual == meta['sr']
        for k, v in meta['channels_schema'].items():
            assert r.channels_schema[k] == v
        assert r.dtype == np.dtype(meta['dtype'])
        assert r.fp_stored == fp
        if 'nodata' in meta['channels_schema']:
            assert r.nodata == meta['channels_schema']['nodata'][0]
        else:
            assert r.nodata == None
        assert len(r) == meta['channel_count']
        assert r.fp == fp
        assert r.mode == 'w'
        assert np.all(r.get_data(channels=slice(None)) == arr)

def test_numpy(meta_numpy):
    meta = meta_numpy
    arrptr = meta_numpy['array'].__array_interface__['data'][0]

    fp = meta['fp']

    ds = Dataset()
    arr = np.add(*fp.meshgrid_raster)
    arr = np.repeat(arr[..., np.newaxis], meta['array'].shape[-1], -1)

    with ds.awrap_numpy_raster(**meta).close as r:
        r.set_data(arr)

        assert r.wkt_stored == meta['sr']
        assert r.wkt_virtual == meta['sr']
        for k, v in meta['channels_schema'].items():
            assert r.channels_schema[k] == tuple(v)
        assert r.dtype == meta['array'].dtype
        assert r.fp_stored == fp
        if 'nodata' in meta['channels_schema']:
            assert r.nodata == meta['channels_schema']['nodata'][0]
        else:
            assert r.nodata == None
        assert len(r) == meta['array'].shape[-1]
        assert r.fp == fp
        assert r.mode == 'w'
        assert np.all(r.get_data(channels=slice(None)) == arr)
        r.array
        assert (
            meta_numpy['array'].__array_interface__['data'][0] ==
            r.array.__array_interface__['data'][0]
        )



import multiprocessing as mp
import multiprocessing.pool
import shutil
import uuid
import tempfile
import functools
import os
import glob
import time
import gc
import threading
import itertools

import numpy as np
import pytest

import buzzard as buzz

def pytest_generate_tests(metafunc):
    if 'pools' in metafunc.fixturenames:
        argvalues = []
        for pval in [
                None,
                'lol',
                mp.pool.ThreadPool(2),
                mp.pool.Pool(2),
        ]:
            # TODO: test with different pools
            # TODO: test with spawn/forks
            argvalues.append(dict(
                io={'io_pool': pval},
                computation={'computation_pool': pval},
                merge={'merge_pool': pval},
                resample={'resample_pool': pval},
            ))

        metafunc.parametrize(
            argnames='pools',
            argvalues=argvalues,
        )

@pytest.fixture
def test_prefix():
    path = os.path.join(tempfile.gettempdir(), 'buzz-ut-' + str(uuid.uuid4()))
    os.makedirs(path)
    yield path
    shutil.rmtree(path)

test_prefix2 = test_prefix

@pytest.fixture(params=[
    (100, 100),
    (99, 99),
    (26, 26),
])
def cache_tiles(request):
    return request.param

def test_(pools, test_prefix, cache_tiles, test_prefix2):
    def _open(**kwargs):
        d = dict(
            fp=fp, dtype='float32', channel_count=2,
            compute_array=functools.partial(_meshgrid_raster_in, reffp=fp),
            cache_dir=test_prefix,
            cache_tiles=cache_tiles,
            **dict(itertools.chain(
                pools['merge'].items(),
                pools['resample'].items(),
                pools['computation'].items(),
                pools['io'].items(),
            ))
        )
        d.update(kwargs)
        return ds.acreate_cached_raster_recipe(**d)

    def _test_get():
        arrs = r.get_data(band=-1)
        assert arrs.shape == tuple(np.r_[fp.shape, 2])
        x, y = arrs[..., 0], arrs[..., 1]
        xref, yref = fp.meshgrid_raster
        assert np.all(x == xref)
        assert np.all(y == yref)

    def _test_resampling(fp):
        arr = r.get_data(band=-1, fp=fp)
        ref = npr.get_data(band=-1, fp=fp)
        assert np.allclose(arr, ref)

    def _corrupt_files(files):
        for path in files:
            with open(path, 'wb') as stream:
                stream.write(b'42')

    print() # debug line
    fp = buzz.Footprint(
        rsize=(100, 100),
        size=(100, 100),
        tl=(1000, 1100),
    )
    compute_same_address_space = (
        type(pools['computation']['computation_pool']) in {str, mp.pool.ThreadPool, type(None)}
    )

    with buzz.Dataset(allow_interpolation=1).close as ds:
        # Create a numpy raster with the same data, useful to compare resampling
        npr = ds.awrap_numpy_raster(fp, np.stack(fp.meshgrid_raster, axis=2).astype('float32'))

        # Test lazyness of cache
        r = _open()
        files = glob.glob(os.path.join(test_prefix, '*.tif'))
        assert len(files) == 0

        # Test get_data results
        _test_get()
        files = glob.glob(os.path.join(test_prefix, '*.tif'))
        assert len(files) > 0
        mtimes0 = {f: os.stat(f).st_mtime for f in files}

        # Test persistence of cache
        # Test get_data results
        r.close()
        r = _open(compute_array=_should_not_be_called)
        _test_get()
        files = glob.glob(os.path.join(test_prefix, '*.tif'))
        assert len(files) > 0
        mtimes1 = {f: os.stat(f).st_mtime for f in files}
        assert mtimes0 == mtimes1

        # Test overwrite parameter
        # Test get_data results
        r.close()
        r = _open(ow=True)
        _test_get()
        files = glob.glob(os.path.join(test_prefix, '*.tif'))
        assert len(files) > 0
        mtimes1 = {f: os.stat(f).st_mtime for f in files}
        assert mtimes0.keys() == mtimes1.keys()
        for k, t0 in mtimes0.items():
            t1 = mtimes1[k]
            assert t0 < t1

        # Test remapping #1 - Interpolation - Fully Inside
        fp_within_upscaled = fp.intersection(fp, scale=fp.scale / 2) & fp.erode(fp.rsemiminoraxis // 4)
        _test_resampling(fp_within_upscaled)

        # Test remapping #2 - Interpolation - Fully Outside
        _test_resampling(fp_within_upscaled.move(fp.br + fp.diagvec))

        # Test remapping #3 - No Interpolation - Fully Outside
        _test_resampling(fp.move(fp.br + fp.diagvec))

        # Test remapping #4 - Interpolation - Both in and out
        _test_resampling(fp_within_upscaled.move(fp.br - fp_within_upscaled.diagvec / 2))

        # Test remapping #5 - No Interpolation - Both in and out
        _test_resampling(fp.move(fp.br - fp.pxvec * fp.rsemiminoraxis))

        # Test remapping #6 - Interpolation - Fully Inside - Tiled
        r.close()
        r = _open(max_resampling_size=20)
        _test_resampling(fp_within_upscaled)

        # Concurrent queries that need a cache file checksum
        r.close()
        r = _open()
        for it in [r.iter_data(fps=[fp], band=-1) for _ in range(10)]:
            next(it)

        # Concurrent queries that need a cache file missing, all but one computation aborted
        # because already launched
        r.close()
        r = _open(ow=True)
        for it in [r.iter_data(fps=[fp], band=-1) for _ in range(10)]:
            next(it)

        # Query garbage collected
        it1 = r.iter_data(fps=[fp] * 2, max_queue_size=1) # 2/2 ready, 1/2 sinked
        it2 = r.iter_data(fps=[fp] * 1, max_queue_size=1) # 1/1 ready, 0/1 sinked
        it3 = r.iter_data(fps=[fp] * 2, max_queue_size=1) # 1/2 ready, 0/2 sinked
        next(it1)
        time.sleep(1/2)

        del it1, it2, it3
        gc.collect()
        time.sleep(1 / 2)
        r.get_data() # This line will reraise any exception from scheduler

        # Raster closing during query
        it1 = r.iter_data(fps=[fp] * 2, max_queue_size=1) # 2/2 ready, 1/2 sinked
        it2 = r.iter_data(fps=[fp] * 1, max_queue_size=1) # 1/1 ready, 0/1 sinked
        it3 = r.iter_data(fps=[fp] * 2, max_queue_size=1) # 1/2 ready, 0/2 sinked
        next(it1)
        time.sleep(1/2)
        # Close Dataset instead of Raster, because Dataset.close is currently blocking

    with buzz.Dataset(allow_interpolation=1).close as ds:
        npr = ds.awrap_numpy_raster(fp, np.stack(fp.meshgrid_raster, axis=2).astype('float32'))

        # Corrupted cache file
        files = glob.glob(os.path.join(test_prefix, '*.tif'))
        mtimes0 = {f: os.stat(f).st_mtime for f in files}
        corrupted_path = files[0]
        _corrupt_files([corrupted_path])
        r = _open()

        r.get_data()
        mtimes1 = {f: os.stat(f).st_mtime for f in files}
        assert mtimes0.keys() == mtimes1.keys()
        for path in files:
            if path == corrupted_path:
                assert mtimes0[path] != mtimes1[path]
            else:
                assert mtimes0[path] == mtimes1[path]

    with buzz.Dataset(allow_interpolation=1).close as ds:
        npr = ds.awrap_numpy_raster(fp, np.stack(fp.meshgrid_raster, axis=2).astype('float32'))

        # In iter_data, the first one(s) don't need cache, the next ones need cache file checking and then recomputation
        _corrupt_files(glob.glob(os.path.join(test_prefix, '*.tif')))
        r = _open()
        fps = [
            fp.move(fp.br + fp.diagvec), # Outside
        ] + [fp] * 12
        arrs = list(r.iter_data(band=-1, fps=fps))
        assert len(arrs) == 13
        for tile, arr in zip(fps, arrs):
            assert np.all(arr == npr.get_data(band=-1, fp=tile))

    with buzz.Dataset(allow_interpolation=1).close as ds:
        npr = ds.awrap_numpy_raster(fp, np.stack(fp.meshgrid_raster, axis=2).astype('float32'))

        # Test channels order versus numpy raster
        r = _open()
        for channels in [
                0, 1, None, slice(None), [0, 1], [1, 0], [1, 0, 1],
        ]:
            assert np.all(r.get_data(channels=channels) == npr.get_data(channels=channels))

    with buzz.Dataset(allow_interpolation=1).close as ds:
        # Derived and primitive rasters not computed
        if compute_same_address_space:
            ac0, ac1 = _AreaCounter(fp), _AreaCounter(fp)
        else:
            ac0, ac1 = None, None
        r0 = _open(
            compute_array=functools.partial(_base_computation, area_counter=ac0, reffp=fp),
            ow=True,
        )
        r1 = _open(
            compute_array=functools.partial(_derived_computation, area_counter=ac1, reffp=fp),
            queue_data_per_primitive={'prim': functools.partial(r0.queue_data, band=-1)},
            cache_dir=test_prefix2,
            ow=True,
        )
        assert len(r0.primitives) == 0
        assert len(r1.primitives) == 1
        assert r1.primitives['prim'] is r0

        r1.get_data()
        if compute_same_address_space:
            ac0.check_done()
            ac1.check_done()
        r0.close()
        r1.close()

        # Derived raster not computed
        if compute_same_address_space:
            ac0, ac1 = _AreaCounter(fp), _AreaCounter(fp)
        else:
            ac0, ac1 = None, None
        r0 = _open(
            compute_array=functools.partial(_base_computation, area_counter=ac0, reffp=fp),
            ow=False,
        )
        r1 = _open(
            compute_array=functools.partial(_derived_computation, area_counter=ac1, reffp=fp),
            queue_data_per_primitive={'prim': functools.partial(r0.queue_data, band=-1)},
            cache_dir=test_prefix2,
            ow=True,
        )
        r1.get_data()
        if compute_same_address_space:
            ac0.check_not_done()
            ac1.check_done()
        r0.close()
        r1.close()

        # Primitive raster not computed
        if compute_same_address_space:
            ac0, ac1 = _AreaCounter(fp), _AreaCounter(fp)
        else:
            ac0, ac1 = None, None
        r0 = _open(
            compute_array=functools.partial(_base_computation, area_counter=ac0, reffp=fp),
            ow=True,
        )
        r1 = _open(
            compute_array=functools.partial(_derived_computation, area_counter=ac1, reffp=fp),
            queue_data_per_primitive={'prim': functools.partial(r0.queue_data, band=-1)},
            cache_dir=test_prefix2,
            ow=False,
        )
        r1.get_data()
        if compute_same_address_space:
            ac0.check_not_done()
            ac1.check_not_done()
        r0.close()
        r1.close()

        # Test computation tiles
        if compute_same_address_space:
            ac0, ac1 = _AreaCounter(fp), _AreaCounter(fp)
        else:
            ac0, ac1 = None, None
        r0 = _open(
            compute_array=functools.partial(_base_computation, area_counter=ac0, reffp=fp),
            computation_tiles=(11, 11),
            ow=True,
        )
        r1 = _open(
            compute_array=functools.partial(_derived_computation, area_counter=ac1, reffp=fp),
            queue_data_per_primitive={'prim': functools.partial(r0.queue_data, band=-1)},
            cache_dir=test_prefix2,
            computation_tiles=(22, 22),
            ow=True,
        )
        r1.get_data()
        if compute_same_address_space:
            ac0.check_done()
            ac1.check_done()
        r0.close()
        r1.close()

        # Several queries, one is dropped, the rest is still working
        r0 = _open(
            compute_array=functools.partial(_base_computation, reffp=fp),
            ow=True,
        )
        r1 = _open(
            compute_array=functools.partial(_derived_computation, reffp=fp),
            queue_data_per_primitive={'prim': functools.partial(r0.queue_data, band=-1)},
            cache_dir=test_prefix2,
            ow=True,
        )
        t = r1.cache_tiles.flatten()
        fps0 = t.tolist() * 2
        fps1 = fps0[::-1]
        fps2 = np.roll(t, t.size // 2).tolist() * 2
        fps3 = fps2[::-1]

        it0 = r1.iter_data(fps=fps0)
        it1 = r1.iter_data(fps=fps1)
        it2 = r1.iter_data(fps=fps2)
        it3 = r1.iter_data(fps=fps3)
        del it1

        assert len(list(it3)) == t.size * 2
        assert len(list(it0)) == t.size * 2
        assert len(list(it2)) == t.size * 2

        r0.close()
        r1.close()

        # Computation function crashes, we catch error in main thread
        r = _open(ow=True, compute_array=_please_crash)
        with pytest.raises(NecessaryCrash):
            r.get_data()

# Tools ***************************************************************************************** **
class _AreaCounter(object):
    def __init__(self, fp):
        self._lock = threading.Lock()
        self._fp = fp
        self._mask = np.zeros(fp.shape, 'uint8')

    def increment(self, fp):
        with self._lock:
            self._mask[fp.slice_in(self._fp)] += 1

    def check_not_done(self):
        assert np.all(self._mask == 0)

    def check_done(self):
        assert np.all(self._mask == 1)

def _base_computation(fp, primitive_fps, primtive_arrays, raster, reffp, area_counter=None):
    if area_counter is not None:
        area_counter.increment(fp)
    x, y = fp.meshgrid_raster_in(reffp)
    return np.stack([x, y], axis=2).astype('float32')

def _derived_computation(fp, primitive_fps, primtive_arrays, raster, reffp, area_counter=None):
    if area_counter is not None:
        area_counter.increment(fp)
    assert fp == primitive_fps['prim']
    x, y = fp.meshgrid_raster_in(reffp)
    return np.stack([x, y], axis=2).astype('float32') * primtive_arrays['prim']

def _meshgrid_raster_in(fp, primitive_fps, primtive_arrays, raster, reffp):
    if raster is not None:
        assert raster.fp == reffp
    x, y = fp.meshgrid_raster_in(reffp)
    return np.stack([x, y], axis=2).astype('float32')

class NecessaryCrash(Exception):
    pass

def _please_crash(fp, primitive_fps, primtive_arrays, raster):
    raise NecessaryCrash()

def _should_not_be_called(*args):
    assert False, _should_not_be_called



import threading

class DatasetRegisterMixin(object):
    """Private mixin for the Dataset class containing subroutines for proxies registration"""

    def __init__(self, **kwargs):
        self._keys_of_source = {}
        self._source_of_key = {}
        self._register_lock = threading.Lock()
        super(DatasetRegisterMixin, self).__init__(**kwargs)

    def _register(self, keys, prox):
        with self._register_lock:
            for key in keys:
                if key in self.__dict__: # pragma: no cover
                    raise ValueError('key `{}` is already bound'.format(key))
            self._keys_of_source[prox] = keys
            for key in keys:
                self._source_of_key[key] = prox
                self.__dict__[key] = prox

    def _unregister(self, source):
        for key in self._keys_of_source[source]:
            del self.__dict__[key]
            del self._source_of_key[key]
        del self._keys_of_source[source]



import os
import uuid
import functools

import numpy as np

from buzzard._actors.message import Msg
from buzzard._actors.pool_job import CacheJobWaiting, PoolJobWorking

create_raster = None # lazy import

class ActorWriter(object):
    """Actor that takes care of writing to disk a cache tile that has been computed and merged."""

    def __init__(self, raster):
        self._raster = raster
        self._alive = True
        io_pool = raster.io_pool
        if io_pool is not None:
            self._waiting_room_address = '/Pool{}/WaitingRoom'.format(id(io_pool))
            self._working_room_address = '/Pool{}/WorkingRoom'.format(id(io_pool))
        self._waiting_jobs = set()
        self._working_jobs = set()
        self.address = '/Raster{}/Writer'.format(self._raster.uid)

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_write_this_array(self, cache_fp, array):
        """Receive message: Please write this array to disk.

        Parameters
        ----------
        cache_fp: Footprint of shape (Y, X)
        array: ndarray of shape (Y, X, C)
        """
        msgs = []

        if self._raster.io_pool is None:
            # No `io_pool` provided by user, perform write operation right now on this thread.
            work = Work(self, cache_fp, array)
            path = work.func()
            msgs += [Msg('CacheSupervisor', 'cache_file_written', cache_fp, path)]
        else:
            # Enqueue job in the `Pool/WaitingRoom` actor
            wait = Wait(self, cache_fp, array)
            self._waiting_jobs.add(wait)
            msgs += [Msg(self._waiting_room_address, 'schedule_job', wait)]
        return msgs

    def receive_token_to_working_room(self, job, token):
        """Receive message: It is your turn to use the `Pool/WorkingRoom` actor

        Parameters
        ----------
        job: Wait
        token: pool_waiting_room._PoolToken
        """
        self._waiting_jobs.remove(job)
        work = Work(self, job.cache_fp, job.array)
        self._working_jobs.add(work)
        return [
            Msg(self._working_room_address, 'launch_job_with_token', work, token)
        ]

    def receive_job_done(self, job, result):
        """Receive message: Writing operation is complete

        Parameters
        ----------
        job: Work
        result: str
            Path to the written file
        """
        self._working_jobs.remove(job)
        return [Msg('CacheSupervisor', 'cache_file_written', job.cache_fp, result)]

    def receive_die(self):
        """Receive message: The raster was killed (collect by gc or closed by user)"""
        assert self._alive
        self._alive = False

        msgs = []
        for job in self._waiting_jobs:
            msgs += [Msg(self._waiting_room_address, 'unschedule_job', job)]
        for job in self._working_jobs:
            msgs += [Msg(self._working_room_address, 'cancel_job', job)]
        self._waiting_jobs.clear()
        self._working_jobs.clear()
        self._raster = None

        return msgs

    # ******************************************************************************************* **

class Wait(CacheJobWaiting):
    """Job to be fed to a PoolWaitingRoom actor"""
    def __init__(self, actor, cache_fp, array):
        self.cache_fp = cache_fp
        self.array = array
        super().__init__(actor.address, actor._raster.uid, self.cache_fp, 2, self.cache_fp)

class Work(PoolJobWorking):
    """Job to be fed to a PoolWorkingRoom actor"""
    def __init__(self, actor, cache_fp, array):
        self.cache_fp = cache_fp

        func = functools.partial(
            _cache_file_write,
            array,
            actor._raster.cache_dir,
            actor._raster.fname_prefix_of_cache_fp(cache_fp),
            '.tif',
            cache_fp,
            {'nodata': actor._raster.nodata},
            actor._raster.wkt_stored,
        )
        actor._raster.debug_mngr.event('object_allocated', func)

        super().__init__(actor.address, func)

def _checksum(fname, buffer_size=512 * 1024, dtype='uint64'):
    # https://github.com/earthcube-lab/buzzard/pull/39/#discussion_r239071556
    dtype = np.dtype(dtype)
    dtypesize = dtype.itemsize
    assert buffer_size % dtypesize == 0
    assert np.issubdtype(dtype, np.unsignedinteger)

    acc = dtype.type(0)
    with open(fname, "rb") as f:
        with np.warnings.catch_warnings():
            np.warnings.filterwarnings('ignore', r'overflow encountered')

            for chunk in iter(lambda: f.read(buffer_size), b""):
                head = np.frombuffer(chunk, dtype, count=len(chunk) // dtypesize)
                head = np.add.reduce(head, dtype=dtype, initial=acc)
                acc += head

                tailsize = len(chunk) % dtypesize
                if tailsize > 0:
                    # This should only be needed for file's tail
                    tail = chunk[-tailsize:] + b'\0' * (dtypesize - tailsize)
                    tail = np.frombuffer(tail, dtype)
                    acc += tail
        return '{:016x}'.format(acc.item())

def _cache_file_write(array,
                      dir_path, filename_prefix, filename_suffix,
                      cache_fp, channels_schema, sr):
    """Write this ndarray to disk.

    It can't use the dataset's activation pool because the file must be closed after
    writing to:
    1. flush to disk
    2. checksum
    3. be renamed

    Parameters
    ----------
    array: ndarray of shape (Y, X, C)
        What to write in the cache file
    dir_path: str
        Directory where to create the file
    filename_prefix: str
        First third of the file name
    filename_suffix: str
        Last thirf of the file name
    cache_fp: Footprint of shape (Y, X)
        Footprint of the file
    channels_schema: dict
        Band schema given by user when creating the cached recipe
    sr: str or None
        Spatial reference given by user when creating the cached recipe
    """
    # Step 0. Lazily import buzzard to avoid circular dependencies
    global create_raster
    if create_raster is None:
        from buzzard import create_raster

    # Step 1. Create/close file with a temporary name
    src_path = os.path.join(
        dir_path, 'tmp_' + filename_prefix + str(uuid.uuid4()) + filename_suffix
    )

    # TODO: Allow the user to choose the type and options of cache files
    # TODO: Use driver-object allocator
    options = [
        "TILED=YES",
        "BLOCKXSIZE=256", "BLOCKYSIZE=256",
        "SPARSE_OK=TRUE",
    ]
    assert array.ndim == 3
    with create_raster(src_path, cache_fp, array.dtype, array.shape[-1], channels_schema,
                       sr=sr, options=options).close as r:
        r.set_data(array, channels=None)

    # Step 2. checksum hash file
    checksum = _checksum(src_path)

    # Step 3. move file to its final location
    dst_path = os.path.join(dir_path, filename_prefix + '_' + checksum + filename_suffix)

    # TODO: Undefined if it exists, but it will most likely work
    # TODO: chmod to remove write access?
    os.rename(src_path, dst_path)

    return dst_path




class Priorities(object):
    """Small class that represents a single version of the priorities among tasks. A new one is
    instantiated and broadcasted as soon as the priorities change.

    """
    __slots__ = ['_prio_actor', '_db_version']

    def __init__(self, prio_actor, db_version):
        self._prio_actor = prio_actor
        self._db_version = db_version

    def prio_of_prod_tile(self, qi, prod_idx):
        if self._prio_actor is None:
            return (prod_idx,)
        else:
            assert self._db_version == self._prio_actor.db_version, 'Failed to fetch latest priorities object'
            return self._prio_actor.prio_of_prod_tile(qi, prod_idx)

    def prio_of_cache_tile(self, raster_uid, cache_fp):
        if self._prio_actor is None:
            return (0,)
        else:
            assert self._db_version == self._prio_actor.db_version, 'Failed to fetch latest priorities object'
            return self._prio_actor.prio_of_cache_tile(raster_uid, cache_fp)

dummy_priorities = Priorities(None, None)



"""Conversions between numpy dtypes and gdal GDTs

All types are not available on all platforms, hence the `_eval_filter_dict_key({` declarations.
"""

from osgeo import gdal
import numpy as np

DTYPE_OF_NAME = {np.dtype(v).name: np.dtype(v) for v in np.typeDict.values()}

def _eval_filter_dict_key(d):
    return {
        DTYPE_OF_NAME[k]: v
        for (k, v) in d.items()
        if k in DTYPE_OF_NAME
    }

def _eval_filter_dict_value(d):
    return {
        k: DTYPE_OF_NAME[v]
        for (k, v) in d.items()
        if v in DTYPE_OF_NAME
    }

# DTYPE -> GDT CONVERSIONS ********** **
_GDT_OF_DTYPE_EQUIV = _eval_filter_dict_key({
    'bool': gdal.GDT_Byte,
    'uint8': gdal.GDT_Byte,

    'float32': gdal.GDT_Float32,
    'float64': gdal.GDT_Float64,
    'int16': gdal.GDT_Int16,
    'int32': gdal.GDT_Int32,
    'uint16': gdal.GDT_UInt16,
    'uint32': gdal.GDT_UInt32,

    'complex128': gdal.GDT_CFloat64,
    'complex64': gdal.GDT_CFloat32,
})

_GDT_OF_DTYPE_UPCAST = _eval_filter_dict_key({
    'float16': gdal.GDT_Float32, # 16 to 32 bits
    'int8': gdal.GDT_Int16, # 8 to 16 bits
})

_GDT_OF_DTYPE_DOWNCAST = _eval_filter_dict_key({
    'float128': gdal.GDT_Float64, # 128 to 64 bits
    'complex256': gdal.GDT_CFloat64, # 128 to 64 bits
    'uint64': gdal.GDT_CFloat64, # 64 to 53 bits
    'int64': gdal.GDT_CFloat64, # 64 to 53 bits
})

_GDT_OF_DTYPE_NOEQUIV = set([
    np.dtype(np.datetime64),
    np.dtype(np.object_),
    np.dtype(np.str_),
    np.dtype(np.timedelta64),
    np.dtype(np.void),
])

# GDT -> DTYPE CONVERSIONS ********** **
_DTYPE_OF_GDT_EQUIV = _eval_filter_dict_value({
    gdal.GDT_Byte: 'uint8',
    gdal.GDT_Int16: 'int16',
    gdal.GDT_Int32: 'int32',
    gdal.GDT_UInt16: 'uint16',
    gdal.GDT_UInt32: 'uint32',
    gdal.GDT_Float32: 'float32',
    gdal.GDT_Float64: 'float64',
    gdal.GDT_CFloat32: 'complex64',
    gdal.GDT_CFloat64: 'complex128',
})

_DTYPE_OF_GDT_UPCAST = _eval_filter_dict_value({
    gdal.GDT_CInt16: 'complex64', # 16 to 24 bits
    gdal.GDT_CInt32: 'complex128', # 32 to 53 bits
})

_DTYPE_OF_GDT_DOWNCAST = dict()

_DTYPE_OF_GDT_NOEQUIV = set()

# STRINGIFICATION ******************* **
def _str_of_gdt(gdt):
    return {
        gdal.GDT_Byte: 'GDT_Byte',
        gdal.GDT_Int16: 'GDT_Int16',
        gdal.GDT_Int32: 'GDT_Int32',
        gdal.GDT_UInt16: 'GDT_UInt16',
        gdal.GDT_UInt32: 'GDT_UInt32',
        gdal.GDT_Float32: 'GDT_Float32',
        gdal.GDT_Float64: 'GDT_Float64',
        gdal.GDT_CFloat32: 'GDT_CFloat32',
        gdal.GDT_CFloat64: 'GDT_CFloat64',
        gdal.GDT_CInt16: 'GDT_CInt16',
        gdal.GDT_CInt32: 'GDT_CInt32',
    }[gdt]

def _str_of_dtype(dtype):
    return str(dtype)

# NORMALIZATION ********************* **
def _to_gdt_opt(gdt):
    """Unchecked GDT -> {None | GDT}"""
    if gdt in _DTYPE_OF_GDT_EQUIV:
        return gdt
    if gdt in _DTYPE_OF_GDT_UPCAST:
        return gdt
    if gdt in _DTYPE_OF_GDT_DOWNCAST:
        return gdt
    if gdt in _DTYPE_OF_GDT_NOEQUIV:
        return gdt
    return None

def _to_dtype_opt(dtype):
    """Unchecked dtype -> {None | dtype}"""
    try:
        dtype = np.dtype(dtype)
    except TypeError:
        return None
    else:
        if dtype in _GDT_OF_DTYPE_EQUIV:
            return dtype
        if dtype in _GDT_OF_DTYPE_UPCAST:
            return dtype
        if dtype in _GDT_OF_DTYPE_DOWNCAST:
            return dtype
        if dtype in _GDT_OF_DTYPE_NOEQUIV:
            return dtype
        return None

# PUBLIC **************************** **
def gdt_of_dtype_equiv(dtype):
    """
    Convert a dtype (numpy data type) to an equivalent GDT (GDAL type).
    If impossible an exception is raised.
    """
    param = dtype
    dtype = _to_dtype_opt(dtype)
    if dtype is None:
        raise ValueError('`%s` is not a dtype' % param)
    gdt = _GDT_OF_DTYPE_EQUIV.get(dtype)
    if gdt is None:
        raise ValueError('`%s` has no equivalent gdt' % _str_of_dtype(dtype))
    return gdt

def gdt_of_dtype_upcast(dtype):
    """
    Convert a dtype (numpy data type) to an equivalent or larger GDT (GDAL type).
    If impossible an exception is raised.
    """
    param = dtype
    dtype = _to_dtype_opt(dtype)
    if dtype is None:
        raise ValueError('`%s` is not a dtype' % param)
    gdt = _GDT_OF_DTYPE_EQUIV.get(dtype) or _GDT_OF_DTYPE_UPCAST.get(dtype)
    if gdt is None:
        raise ValueError('`%s` has no equivalent or upcast gdt' % _str_of_dtype(dtype))
    return gdt

def gdt_of_dtype_downcast(dtype):
    """
    Convert a dtype (numpy data type) to an equivalent or larger or smaller GDT (GDAL type).
    If impossible an exception is raised.
    """
    param = dtype
    dtype = _to_dtype_opt(dtype)
    if dtype is None:
        raise ValueError('`%s` is not a dtype' % param)
    gdt = (
        _GDT_OF_DTYPE_EQUIV.get(dtype) or
        _GDT_OF_DTYPE_UPCAST.get(dtype) or
        _GDT_OF_DTYPE_DOWNCAST.get(dtype)
    )
    if gdt is None:
        raise ValueError('`%s` has no equivalent or upcast or downcast gdt' % _str_of_dtype(dtype))
    return gdt

def dtype_of_gdt_equiv(gdt):
    """
    Convert a GDT (GDAL type) to an equivalent dtype (numpy data type).
    If impossible an exception is raised.
    """
    param = gdt
    gdt = _to_gdt_opt(gdt)
    if gdt is None:
        raise ValueError('`%s` is not a gdt' % param)
    dtype = _DTYPE_OF_GDT_EQUIV.get(gdt)
    if dtype is None:
        raise ValueError('`%s` has no equivalent dtype' % _str_of_gdt(gdt))
    return dtype

def dtype_of_gdt_upcast(gdt):
    """
    Convert a GDT (GDAL type) to an equivalent or larger dtype (numpy data type).
    If impossible an exception is raised.
    """
    param = gdt
    gdt = _to_gdt_opt(gdt)
    if gdt is None:
        raise ValueError('`%s` is not a gdt' % param)
    dtype = _DTYPE_OF_GDT_EQUIV.get(gdt)
    if dtype is None:
        dtype = _DTYPE_OF_GDT_UPCAST.get(gdt)
    if dtype is None:
        raise ValueError('`%s` has no equivalent or upcast dtype' % _str_of_gdt(gdt))
    return dtype

def dtype_of_gdt_downcast(gdt):
    """
    Convert a GDT (GDAL type) to an equivalent or larger or smaller dtype (numpy data type).
    If impossible an exception is raised.
    """
    param = gdt
    gdt = _to_gdt_opt(gdt)
    if gdt is None:
        raise ValueError('`%s` is not a gdt' % param)
    dtype = _DTYPE_OF_GDT_EQUIV.get(gdt)
    if dtype is None:
        dtype = _DTYPE_OF_GDT_UPCAST.get(gdt)
    if dtype is None:
        dtype = _DTYPE_OF_GDT_DOWNCAST.get(gdt)
    if dtype is None:
        raise ValueError('`%s` has no equivalent or upcast or downcast dtype' % _str_of_gdt(gdt))
    return dtype

def gdt_of_any_equiv(obj):
    """
    Convert a GDT (GDAL type) or a dtype (numpy data type) to an equivalent GDT.
    If impossible an exception is raised.
    """
    gdt = _to_gdt_opt(obj)
    if gdt is not None:
        return gdt
    dtype = _to_dtype_opt(obj)
    if dtype is not None:
        return gdt_of_dtype_equiv(obj)
    raise ValueError('Unknown type')

def gdt_of_any_upcast(obj):
    """
    Convert a GDT (GDAL type) or a dtype (numpy data type) to an equivalent or larger GDT.
    If impossible an exception is raised.
    """
    gdt = _to_gdt_opt(obj)
    if gdt is not None:
        return gdt
    dtype = _to_dtype_opt(obj)
    if dtype is not None:
        return gdt_of_dtype_upcast(obj)
    raise ValueError('Unknown type')

def gdt_of_any_downcast(obj):
    """
    Convert a GDT (GDAL type) or a dtype (numpy data type) to an equivalent or larger or smaller
    GDT.
    If impossible an exception is raised.
    """
    gdt = _to_gdt_opt(obj)
    if gdt is not None:
        return gdt
    dtype = _to_dtype_opt(obj)
    if dtype is not None:
        return gdt_of_dtype_downcast(obj)
    raise ValueError('Unknown type')

def dtype_of_any_equiv(obj):
    """
    Convert a GDT (GDAL type) or a dtype (numpy data type) to an equivalent dtype.
    If impossible an exception is raised.
    """
    dtype = _to_dtype_opt(obj)
    if dtype is not None:
        return dtype
    gdt = _to_gdt_opt(obj)
    if gdt is not None:
        return dtype_of_gdt_equiv(obj)
    raise ValueError('Unknown type')

def dtype_of_any_upcast(obj):
    """
    Convert a GDT (GDAL type) or a dtype (numpy data type) to an equivalent or larger dtype.
    If impossible an exception is raised.
    """
    dtype = _to_dtype_opt(obj)
    if dtype is not None:
        return dtype
    gdt = _to_gdt_opt(obj)
    if gdt is not None:
        return dtype_of_gdt_upcast(obj)
    raise ValueError('Unknown type')

def dtype_of_any_downcast(obj):
    """
    Convert a GDT (GDAL type) or a dtype (numpy data type) to an equivalent or larger or smaller
    dtype.
    If impossible an exception is raised.
    """
    dtype = _to_dtype_opt(obj)
    if dtype is not None:
        return dtype
    gdt = _to_gdt_opt(obj)
    if gdt is not None:
        return dtype_of_gdt_downcast(obj)
    raise ValueError('Unknown type')



import collections
import threading
import contextlib

from buzzard._tools import MultiOrderedDict

_ERR_FMT = 'Dataset is configured for a maximum of {} simultaneous active driver objects \
but there are already {} idle objects and {} used objects'

class BackDatasetActivationPoolMixin(object):
    """Private mixin for the Dataset class containing subroutines for proxies' driver
    objects pooling"""

    def __init__(self, max_active, **kwargs):
        self.max_active = max_active
        self._ap_lock = threading.Lock()
        self._ap_idle = MultiOrderedDict()
        self._ap_used = collections.Counter()
        super(BackDatasetActivationPoolMixin, self).__init__(**kwargs)

    def activate(self, uid, allocator):
        """Make sure at least one driver object is idle or used for uid"""
        with self._ap_lock:
            if self._ap_used[uid] == 0 and uid not in self._ap_idle:
                self._ensure_one_slot()
                self._ap_idle.push_front(uid, allocator())

    def deactivate(self, uid):
        """Flush all occurrences of uid from _ap_idle. Raises an exception if uid is in _ap_used
        """
        with self._ap_lock:
            if self._ap_used[uid] > 0:
                raise RuntimeError('Attempting to deactivate a source currently used')
            self._ap_idle.pop_all_occurrences(uid)

    def deactivate_many(self, uid_set):
        # TODO idea: allow recursive uids to group activated rasters and allow group deactivation
        if len(uid_set) == 0:
            return
        with self._ap_lock:
            being_used = uid_set & self._ap_used.keys()
            if being_used:
                raise RuntimeError('Attempting to deactivate {} source currently used'.format(
                    len(being_used)
                ))
            idle = self._ap_idle & uid_set
            for uid in idle:
                self._ap_idle.pop_all_occurrences(uid)

    def used_count(self, uid=None):
        """Count how many driver objects exist for uid"""
        with self._ap_lock:
            if uid is None:
                return sum(self._ap_used.values())
            else:
                return self._ap_used[uid]

    def idle_count(self, uid=None):
        """Count how many driver objects exist for uid"""
        with self._ap_lock:
            if uid is None:
                return len(self._ap_idle)
            else:
                return self._ap_idle.count(uid)

    def active_count(self, uid=None):
        """Count how many driver objects exist for uid"""
        with self._ap_lock:
            if uid is None:
                return len(self._ap_idle) + sum(self._ap_used.values())
            else:
                return self._ap_idle.count(uid) + self._ap_used[uid]

    def acquire_driver_object(self, uid, allocator):
        """Return a context manager to acquire a driver object

        Example
        -------
        >>> with back_ds.acquire(uid) as gdal_obj:
        ...     pass
        """
        @contextlib.contextmanager
        def _acquire():
            with self._ap_lock:
                if uid in self._ap_idle:
                    obj = self._ap_idle.pop_first_occurrence(uid)
                    allocate = False
                else:
                    self._ensure_one_slot()
                    allocate = True
                self._ap_used[uid] += 1

            if allocate:
                try:
                    obj = allocator()
                except:
                    with self._ap_lock:
                        self._ap_used[uid] -= 1
                    raise

            try:
                yield obj
            finally:
                with self._ap_lock:
                    self._ap_used[uid] -= 1
                    assert self._ap_used[uid] >= 0
                    self._ap_idle.push_front(uid, obj)
                    if self._ap_used[uid] == 0:
                        del self._ap_used[uid]

        return _acquire()

    def _ensure_one_slot(self):
        total = sum(self._ap_used.values()) + len(self._ap_idle)
        assert total <= self.max_active
        if total == self.max_active:
            if len(self._ap_idle) == 0:
                raise RuntimeError(_ERR_FMT.format(
                    self.max_active,
                    len(self._ap_idle),
                    sum(self._ap_used.values()),
                ))
            self._ap_idle.pop_back()



""">>> help(TileMixin)"""

import numpy as np

class TileMixin(object):
    """Private mixin for the Footprint class containing tiling subroutines"""

    _TILE_BOUNDARY_EFFECTS = set(['extend', 'exclude', 'overlap', 'shrink', 'exception'])
    _TILE_OCCURRENCE_BOUNDARY_EFFECTS = set(['extend', 'exception'])
    _TILE_BOUNDARY_EFFECT_LOCI = set(['br', 'tr', 'tl', 'bl'])

    @staticmethod
    def _details_of_tiling_direction(tile_size, overlap_size, raster_size):
        """Compute tiling operation details"""
        increment = tile_size - overlap_size
        if tile_size > raster_size:
            count = 0
        else:
            count = 1 + (raster_size - tile_size) // increment
        gap = raster_size - tile_size - max(0, count - 1) * increment

        def _gen():
            val = 0
            for _ in range(count):
                yield val
                val += increment
        return gap, _gen()

    def _tile_extend_deltax_gen(self, sizex, overlapx):
        gap, gen = self._details_of_tiling_direction(sizex, overlapx, self.rsizex)
        if gap < 0:
            yield 0, sizex
        else:
            for val in gen:
                yield val, sizex
            if gap != 0:
                yield self.rsizex - gap - overlapx, sizex

    def _tile_extend_deltay_gen(self, sizey, overlapy):
        gap, gen = self._details_of_tiling_direction(sizey, overlapy, self.rsizey)
        if gap < 0:
            yield 0, sizey
        else:
            for val in gen:
                yield val, sizey
            if gap != 0:
                yield self.rsizey - gap - overlapy, sizey

    def _tile_exclude_deltax_gen(self, sizex, overlapx):
        _, gen = self._details_of_tiling_direction(sizex, overlapx, self.rsizex)
        for val in gen:
            yield val, sizex

    def _tile_exclude_deltay_gen(self, sizey, overlapy):
        _, gen = self._details_of_tiling_direction(sizey, overlapy, self.rsizey)
        for val in gen:
            yield val, sizey

    def _tile_overlap_deltax_gen(self, sizex, overlapx):
        gap, gen = self._details_of_tiling_direction(sizex, overlapx, self.rsizex)
        if gap < 0:
            raise ValueError(
                'Cannot apply boundary_effect=overlap with a tile(%s) bigger than source(%s)' % (
                    sizex, self.rsizex
                ))
        else:
            for val in gen:
                yield val, sizex
            if gap != 0:
                yield self.rsizex - sizex, sizex

    def _tile_overlap_deltay_gen(self, sizey, overlapy):
        gap, gen = self._details_of_tiling_direction(sizey, overlapy, self.rsizey)
        if gap < 0:
            raise ValueError(
                'Cannot apply boundary_effect=overlap with a tile(%s) bigger than source(%s)' % (
                    sizey, self.rsizey
                ))
        else:
            for val in gen:
                yield val, sizey
            if gap != 0:
                yield self.rsizey - sizey, sizey

    def _tile_raise_deltax_gen(self, sizex, overlapx):
        gap, gen = self._details_of_tiling_direction(sizex, overlapx, self.rsizex)
        if gap != 0:
            raise ValueError(
                ('There is a gap of %d pixel in the x direction, ' +
                 '`gap:%d %% (sizex:%d - overlapx:%d) == 0` was required') % (

                     (gap, gap, sizex, overlapx)))
        for val in gen:
            yield val, sizex

    def _tile_raise_deltay_gen(self, sizey, overlapy):
        gap, gen = self._details_of_tiling_direction(sizey, overlapy, self.rsizey)
        if gap != 0:
            raise ValueError(
                ('There is a gap of %d pixel in the y direction, ' +
                 '`gap:%d %% (sizey:%d - overlapy:%d) == 0` was required') % (
                     (gap, gap, sizey, overlapy)))
        for val in gen:
            yield val, sizey

    def _tile_shrink_deltax_gen(self, sizex, overlapx):
        gap, gen = self._details_of_tiling_direction(sizex, overlapx, self.rsizex)
        if gap < 0:
            yield 0, self.rsizex
        else:
            for val in gen:
                yield val, sizex
            if gap != 0:
                yield self.rsizex - gap - overlapx, gap + overlapx

    def _tile_shrink_deltay_gen(self, sizey, overlapy):
        gap, gen = self._details_of_tiling_direction(sizey, overlapy, self.rsizey)
        if gap < 0:
            yield 0, self.rsizey
        else:
            for val in gen:
                yield val, sizey
            if gap != 0:
                yield self.rsizey - gap - overlapy, gap + overlapy

    def _tile_unsafe(self, size, overlapx, overlapy, boundary_effect, boundary_effect_locus):
        if boundary_effect == 'extend':
            gen_xinfo = self._tile_extend_deltax_gen(size[0], overlapx)
            gen_yinfo = self._tile_extend_deltay_gen(size[1], overlapy)
        elif boundary_effect == 'exclude':
            gen_xinfo = self._tile_exclude_deltax_gen(size[0], overlapx)
            gen_yinfo = self._tile_exclude_deltay_gen(size[1], overlapy)
        elif boundary_effect == 'overlap':
            gen_xinfo = self._tile_overlap_deltax_gen(size[0], overlapx)
            gen_yinfo = self._tile_overlap_deltay_gen(size[1], overlapy)
        elif boundary_effect == 'shrink':
            gen_xinfo = self._tile_shrink_deltax_gen(size[0], overlapx)
            gen_yinfo = self._tile_shrink_deltay_gen(size[1], overlapy)
        elif boundary_effect == 'exception':
            gen_xinfo = self._tile_raise_deltax_gen(size[0], overlapx)
            gen_yinfo = self._tile_raise_deltay_gen(size[1], overlapy)
        else:
            assert False # pragma: no cover

        if boundary_effect_locus == 'br':
            origin = self.tl
            direction = np.array([+1, +1], dtype='int')
        elif boundary_effect_locus == 'tr':
            origin = self.bl
            direction = np.array([+1, -1], dtype='int')
        elif boundary_effect_locus == 'tl':
            origin = self.br
            direction = np.array([-1, -1], dtype='int')
        elif boundary_effect_locus == 'bl':
            origin = self.tr
            direction = np.array([-1, +1], dtype='int')
        else:
            assert False # pragma: no cover

        horiz_vec = self.pxlrvec * direction[0]
        vert_vec = self.pxtbvec * direction[1]

        def _footprint_of_deltacoords(y, x, sizex, sizey):
            rsize = np.asarray([sizex, sizey])
            gt = self.gt
            tl = horiz_vec * x + vert_vec * y + origin
            tl -= rsize * (direction == -1) * (1, -1) # I don't get this line :'(
            gt[0] = tl[0]
            gt[3] = tl[1]
            return self.__class__(
                gt=gt,
                rsize=rsize,
            )

        infoxs = list(gen_xinfo)
        infoys = list(gen_yinfo)
        deltaxs, deltays = np.meshgrid(
            [deltax for (deltax, _) in infoxs],
            [deltay for (deltay, _) in infoys],
        )
        sizexs, sizeys = np.meshgrid(
            [sizex for (_, sizex) in infoxs],
            [sizey for (_, sizey) in infoys],
        )
        tiles = np.asarray(list(map(
            _footprint_of_deltacoords,
            deltays.flatten(),
            deltaxs.flatten(),
            sizexs.flatten(),
            sizeys.flatten(),
        )), dtype=object)
        if tiles.size > 0:
            tiles = tiles.reshape(len(infoys), len(infoxs))
        if direction[0] == -1:
            tiles = np.fliplr(tiles)
        if direction[1] == -1:
            tiles = np.flipud(tiles)
        return tiles



"""Tile set creation for Footprint tests"""

from __future__ import division, print_function
import itertools
import attrdict

import numpy as np

from buzzard import Footprint

ALL_LETTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvw"


def make_tile_set(width, reso, tilevec=(1, -10)):
    """
    Tile set creation for Footprint tests
    In a grid of width by width, create a Footprint for all rectangles.

    # *********************************************************************** **
    Ex with=3, reso=(1, -1), tilevec=(1, -10):
    A B C
    D E F
    G H I

    Footprint list (36):
    size 1: A, B, C, D, E, F, G, H, I (9)
    size 2 horizontal: AB, BC, DE, EF, GH, HI (6)
    size 2 vertical: AD, DG, BE, EH, CF, FI (6)
    size 3: AC, DF, GI, AG, BH, CI (6)
    size 4: AE, BF, DH, EI (4)
    size 6: AH, BI, AF, DI (4)
    size 9: AI (1)

    topleft coords:
    A: 0, 30
    B: 1, 30
    C: 2, 30
    D: 0, 20
    E: 1, 20
    F: 2, 20
    G: 0, 10
    H: 1, 10
    I: 2, 10

    # *********************************************************************** **
    All sizes

    1 *********************************************************************** **
    len(fps) = 1
    A
    2 *********************************************************************** **
    len(fps) = 9
    A B
    C D
    3 *********************************************************************** **
    len(fps) = 36
    A B C
    D E F
    G H I
    4 *********************************************************************** **
    len(fps) = 100
    A B C D
    E F G H
    I J K L
    M N O P
    5 *********************************************************************** **
    len(fps) = 225
    A B C D E
    F G H I J
    K L M N O
    P Q R S T
    U V W X Y
    6 *********************************************************************** **
    len(fps) = 441
    A B C D E F
    G H I J K L
    M N O P Q R
    S T U V W X
    Y Z a b c d
    e f g h i j
    7 *********************************************************************** **
    len(fps) = 784
    A B C D E F G
    H I J K L M N
    O P Q R S T U
    V W X Y Z a b
    c d e f g h i
    j k l m n o p
    q r s t u v w
    """
    assert width <= 7
    count = int(width ** 2)
    letters = ALL_LETTERS[0:count]
    reso = np.asarray(reso)
    tilevec = np.asarray(tilevec)

    tlx = 0
    tly = -tilevec[1] * width
    tlxs, tlys = np.meshgrid(
        np.arange(tlx, tilevec[0] * width, tilevec[0]),
        np.arange(tly, tilevec[1] * width, tilevec[1]),
    )

    tl_of_letter = {
        letter: np.array((tlx, tly), dtype=int)
        for letter, tlx, tly in zip(letters, tlxs.flatten(), tlys.flatten())
    }
    br_of_letter = {
        letter: np.array((tlx, tly), dtype=int) + tilevec
        for letter, tlx, tly in zip(letters, tlxs.flatten(), tlys.flatten())
    }

    combos = [
        ''.join([a, b])
        for a, b in itertools.combinations(letters, 2)
        if tl_of_letter[a][0] <= tl_of_letter[b][0] and tl_of_letter[a][1] >= tl_of_letter[b][1]
    ] + list(letters)

    def _footprint_of_letters(letters):
        tl = tl_of_letter[letters[0]]
        br = br_of_letter[letters[-1]]
        diagvec = br - tl
        fp = Footprint(
            tl=tl, size=np.abs(diagvec), rsize=(diagvec / reso)
        )
        return fp
    fps = attrdict.AttrDict({combo: _footprint_of_letters(combo) for combo in combos})
    return fps



import contextlib

from buzzard._a_emissary_vector import AEmissaryVector, ABackEmissaryVector
from buzzard._a_gdal_vector import ABackGDALVector
from buzzard._tools import conv

class GDALMemoryVector(AEmissaryVector):
    """Concrete class defining the behavior of a GDAL raster using the "Memory" driver

    >>> help(Dataset.create_vector)

    Features Defined
    ----------------
    None
    """

    def __init__(self, ds, allocator, open_options):
        back = BackGDALMemoryVector(
            ds._back, allocator, open_options,
        )
        super(GDALMemoryVector, self).__init__(ds=ds, back=back)

class BackGDALMemoryVector(ABackEmissaryVector, ABackGDALVector):
    """Implementation of GDALMemoryVector"""

    def __init__(self, back_ds, allocator, open_options):
        # gdal_ds, lyr = self.create_file('', geometry, fields, layer, 'Memory', open_options, sr)

        gdal_ds, lyr = allocator()
        self._gdal_ds = gdal_ds
        self._lyr = lyr

        rect = None
        if lyr is not None:
            rect = lyr.GetExtent()

        path = gdal_ds.GetDescription()
        driver = gdal_ds.GetDriver().ShortName
        sr = lyr.GetSpatialRef()
        if sr is None:
            wkt_stored = None
        else:
            wkt_stored = sr.ExportToWkt()
        fields = BackGDALMemoryVector._fields_of_lyr(lyr)
        type = conv.str_of_wkbgeom(lyr.GetGeomType())
        layer = lyr.GetName()

        super(BackGDALMemoryVector, self).__init__(
            back_ds=back_ds,
            wkt_stored=wkt_stored,
            mode='w',
            driver=driver,
            open_options=open_options,
            path=path,
            layer=layer,
            fields=fields,
            rect=rect,
            type=type
        )

        self._type_of_field_index = [
            conv.type_of_oftstr(field['type'])
            for field in self.fields
        ]

    @contextlib.contextmanager
    def acquire_driver_object(self):
        yield self._gdal_ds, self._lyr

    def delete(self): # pragma: no cover
        raise NotImplementedError('GDAL Memory driver does no allow deletion, use `close`')

    def close(self):
        super(BackGDALMemoryVector, self).close()
        del self._lyr
        del self._gdal_ds



# Waiting *************************************************************************************** **
class PoolJobWaiting:
    """Base class of all waiting jobs.

    A waiting job has a priority, it waits for a token in a PoolWaitingRoom with other waiting jobs.
    This token allows a waiting job to become a working job, and go to the PoolWorkingRoom to
    get some computations done.
    """
    def __init__(self, sender_address):
        self.sender_address = sender_address

class MaxPrioJobWaiting(PoolJobWaiting):
    pass

class ProductionJobWaiting(PoolJobWaiting):
    def __init__(self, sender_address, qi, prod_idx, action_priority, fp):
        super().__init__(sender_address)
        self.fp = fp
        self.qi = qi
        self.prod_idx = prod_idx
        self.action_priority = action_priority

class CacheJobWaiting(PoolJobWaiting):
    def __init__(self, sender_address, raster_uid, cache_fp, action_priority, fp):
        super().__init__(sender_address)
        self.fp = fp
        self.raster_uid = raster_uid
        self.cache_fp = cache_fp
        self.action_priority = action_priority

# Working *************************************************************************************** **
class PoolJobWorking(object):
    """Base class of all working jobs.

    A working job paired with a token from a PoolWaitingRoom can be fed to a PoolWorkingRoom to
    compute things on the wrapped thread/process pool.
    """
    __slots__ = ['sender_address', 'func']

    def __init__(self, sender_address, func):
        self.sender_address = sender_address
        self.func = func



import logging

from buzzard._actors.message import Msg, DroppableMsg, AgingMsg
from buzzard._actors.cached.query_infos import CachedQueryInfos

LOGGER = logging.getLogger(__name__)

class ActorQueriesHandler(object):
    """Actor that takes care of a raster's queries lifetime"""

    def __init__(self, raster):
        """
        Parameter
        ---------
        raster: _a_recipe_raster.ABackRecipeRaster
        """
        self._raster = raster
        self._queries = {}
        self._alive = True
        self.address = '/Raster{}/QueriesHandler'.format(self._raster.uid)

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def ext_receive_new_query(self, queue_wref, max_queue_size, produce_fps,
                              channel_ids, is_flat, dst_nodata, interpolation, parent_uid,
                              key_in_parent):
        """Receive message sent by something else than an actor, still treated synchronously: There
        is a new query.

        Parameters
        ----------
        queue_wref: weakref.ref of queue.Queue
           Queue returned by the underlying `queue_data` (or behind a `(get|iter)_data`).
        max_queue_size: int
           Max queue size of the queue returned by the underlying `queue_data`
           (or behind a `(get|iter)_data`).
        produce_fps: sequence of Footprint
           Parameter of the underlying `(get|iter|queue)_data`
        channel_ids: sequence of int
           Parameter of the underlying `(get|iter|queue)_data`
        is_flat: bool
           Parameter of the underlying `(get|iter|queue)_data`
        dst_nodata: nbr
           Parameter of the underlying `(get|iter|queue)_data`
        interpolation: str
           Parameter of the underlying `(get|iter|queue)_data`
        parent_uid: None or uuid.UUID4
           uuid of parent raster
           if None: This query comes directly from the user
           else: The id of the parent that issued the query
        key_in_parent: None or object
           identity of this query in the parent query
           if None: This query comes directly from the user
           else: This query was issued by another raster
        """
        msgs = []

        qi = CachedQueryInfos(
            self._raster, produce_fps,
            channel_ids, is_flat, dst_nodata, interpolation,
            max_queue_size,
            parent_uid, key_in_parent,
        )
        self._raster.debug_mngr.event('object_allocated', qi)

        q = _Query(queue_wref)
        self._queries[qi] = q
        msgs += [
            Msg('ProductionGate', 'make_those_arrays', qi),
        ]
        if len(qi.list_of_cache_fp) > 0:
            msgs += [Msg('CacheSupervisor', 'make_those_cache_files_available', qi)]

        return msgs

    def ext_receive_nothing(self):
        """Receive message sent by something else than an actor, still treated synchronously: What's
        up?
        Was an output queue sinked?
        Was an output queue collected by gc?
        """
        msgs = []

        killed_queries = []
        for qi, q in self._queries.items():
            queue = q.queue_wref()
            if queue is None:
                killed_queries.append(qi)
            else:
                new_queue_size = queue.qsize()
                assert new_queue_size <= q.queue_size, "Don't put data in that queue..."
                if new_queue_size != q.queue_size:
                    q.queue_size = new_queue_size
                    msgs += [
                        AgingMsg('/Global/GlobalPrioritiesWatcher', 'output_queue_update',
                                 (self._raster.uid, qi), (q.produced_count, q.queue_size)),
                        AgingMsg('ProductionGate', 'output_queue_update',
                                 (qi,), (q.produced_count, q.queue_size)),
                        AgingMsg('ComputationGate1', 'output_queue_update',
                                 (qi,), (q.produced_count, q.queue_size)),
                    ]
            del q

        for qi in killed_queries:
            msgs += self._cancel_query(qi)

        return msgs

    def receive_made_this_array(self, qi, prod_idx, array):
        """Receive message: This array is ready to be sent to the output queue. Just do it in the
        right order.

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        prod_idx: int
        array: np.ndarray
        """
        msgs = []
        q = self._queries[qi]
        assert prod_idx not in q.produce_arrays_dict, 'This array was already computed'
        assert prod_idx >= q.produced_count, 'This array was already sent'
        q.produce_arrays_dict[prod_idx] = array

        # Send arrays ready ****************************************************
        prod_idx = q.produced_count
        queue = q.queue_wref()
        if queue is None:
            # Queue is None (Queue was collected upstream by gc) -> Ignore the problem,
            # `ext_receive_nothing` will be called soon
            pass
        else:
            update = False

            # Put arrays in queue in the right order
            while True:
                if prod_idx not in q.produce_arrays_dict:
                    # Next array is not ready yet
                    break
                array = q.produce_arrays_dict.pop(prod_idx)

                y, x, c = array.shape
                if qi.is_flat and c == 1:
                    array = array.reshape(y, x)

                # The way this is all designed, the system does not start to work on a `prod_idx` if
                # it cannot be inserted in the output queue. It means that the `queue.Full`
                # exception cannot be raised by the following `put_nowait`.
                queue.put_nowait(array)

                q.queue_size += 1
                q.produced_count += 1

                prod_idx  = q.produced_count
                update = True

            if update:
                msgs += [
                    AgingMsg('/Global/GlobalPrioritiesWatcher', 'output_queue_update',
                             (self._raster.uid, qi), (q.produced_count, q.queue_size)),
                    AgingMsg('ProductionGate', 'output_queue_update',
                             (qi,), (q.produced_count, q.queue_size)),
                    AgingMsg('ComputationGate1', 'output_queue_update',
                             (qi,), (q.produced_count, q.queue_size)),
                ]
                if qi.key_in_parent is not None:
                    # Notify the parent raster that a new array was put in the queue
                    # If the parent raster was collected this message is discarded
                    msgs += [DroppableMsg(
                        '/Raster{}/ComputationGate2'.format(qi.parent_uid),
                        'input_queue_update',
                        qi.key_in_parent,
                    )]

            if q.produced_count == qi.produce_count:
                del self._queries[qi]
        del queue

        return msgs

    def receive_die(self):
        """Receive message: The raster was killed"""
        assert self._alive
        self._alive = False

        msgs = []
        for qi in list(self._queries.keys()):
            msgs += self._cancel_query(qi)

        self._queries.clear()
        self._raster = None
        return msgs

    # ******************************************************************************************* **
    def _cancel_query(self, qi):
        q = self._queries.pop(qi)
        assert q.produced_count != qi.produce_count, "This query finished and can't be cancelled"
        LOGGER.warning('Dropping a query with {}/{} arrays produced.'.format(
            q.produced_count,
            qi.produce_count,
        ))
        return [
            Msg('/Global/GlobalPrioritiesWatcher', 'cancel_this_query', self._raster.uid, qi),

            Msg('ProductionGate', 'cancel_this_query', qi),
            Msg('Producer', 'cancel_this_query', qi),
            Msg('Resampler', 'cancel_this_query', qi),
            Msg('CacheExtractor', 'cancel_this_query', qi),
            Msg('Reader', 'cancel_this_query', qi),

            Msg('CacheSupervisor', 'cancel_this_query', qi),
            Msg('ComputationGate1', 'cancel_this_query', qi),
            Msg('ComputationGate2', 'cancel_this_query', qi),
            Msg('Computer', 'cancel_this_query', qi),
        ]

    # ******************************************************************************************* **

class _Query(object):
    def __init__(self, queue_wref):
        self.queue_wref = queue_wref
        self.produce_arrays_dict = {}
        self.produced_count = 0
        self.queue_size = 0



import threading
import collections
import multiprocessing as mp
import multiprocessing.pool

class PoolsContainer(object):
    """Manages thread/process pools and aliases for a Dataset"""

    def __init__(self):
        self._aliases_per_pool = collections.defaultdict(set)
        self._aliases = {}
        self._managed_pools = set()
        self._lock = threading.Lock()

    def alias(self, key, pool_or_none):
        """Register the given pool under the given key in this Dataset. The key can then be
        used to refer to that pool from within the async raster constructors.

        Parameters
        ----------
        key: hashable (like a string)
            ..
        pool_or_none: multiprocessing.pool.Pool or multiprocessing.pool.ThreadPool or None
            ..
        """
        with self._lock:
            if key in self._aliases: # pragma: no cover
                raise ValueError('Key `{}` is already bound to `{}`'.format(
                    key, self._aliases[key]
                ))
            self._aliases_per_pool[pool_or_none].add(key)
            self._aliases[key] = pool_or_none

    def manage(self, pool):
        """Add the given pool to the list of pools that must be terminated upon Dataset closing.

        Parameters
        ----------
        pool: multiprocessing.pool.Pool or multiprocessing.pool.ThreadPool
            ..

        """
        if not isinstance(pool, (mp.pool.Pool, mp.pool.ThreadPool)): # pragma: no cover
            raise TypeError('Can only manage pools')
        with self._lock:
            self._managed_pools.add(pool)

    def __len__(self):
        """Number of pools registered in this Dataset"""
        with self._lock:
            return len(
                p
                for p in self._aliases_per_pool.keys()
                if p is not None
            )

    def __iter__(self):
        """Generator of pools registered in this Dataset"""
        for p in self._aliases_per_pool.keys():
            if p is not None:
                yield p

    def __getitem__(self, key):
        """Pool or none getter from alias"""
        return self._aliases[key]

    def __contains__(self, obj):
        """Is pool or alias registered in this Dataset"""
        with self._lock:
            return obj in self._aliases or obj in self._aliases_per_pool

    # Private interface with Dataset ********************************************************* **
    def _close(self):
        things_to_join = []
        for pool in self._managed_pools:
            if isinstance(pool, mp.pool.ThreadPool):
                pool.terminate()
                things_to_join.append(pool)
            else:
                things_to_join.append(_create_process_pool_killer(pool))
        for joinable in things_to_join:
            joinable.join()
        self._aliases.clear()
        self._aliases_per_pool.clear()
        self._managed_pools.clear()

    def _normalize_pool_parameter(self, pool_param, param_name):
        if isinstance(pool_param, (mp.pool.Pool, mp.pool.ThreadPool)):
            return pool_param
        if pool_param is None:
            return None
        if not (hasattr(pool_param, '__hash__') and hasattr(pool_param, '__eq__')): # pragma: no cover
            types = [
                'multiprocessing.pool.Pool',
                'multiprocessing.pool.ThreadPool',
                'None', 'hashable',
            ]
            raise TypeError('`{}` parameter should be one of {}'.format(
                param_name, ', '.join(types)
            ))
        with self._lock:
            if pool_param not in self._aliases:
                p = mp.pool.ThreadPool(mp.cpu_count())
                self._aliases[pool_param] = p
                self._aliases_per_pool[p] = [pool_param]
                self._managed_pools.add(p)
        return self._aliases[pool_param]

def _create_process_pool_killer(pp):
    """
    https://stackoverflow.com/questions/42782953/python-concurrent-futures-how-to-make-it-cancelable/45515052#45515052
    """
    def kill_process_pool_from_thread():
        pp.close()
        pp.terminate()
        pp.join()
    t = threading.Thread(target=kill_process_pool_from_thread)
    t.start()
    return t



import numpy as np
import cv2

from buzzard._tools import ANY

_EXN_FORMAT0 = """Illegal remap attempt between two Footprints that do not lie on the same grid.
full raster    -> {src!s}
argument       -> {dst!s}
scales         -> full raster:{src.scale}, argument:{dst.scale}
grids distance -> {tldiff}
"""

_EXN_FORMAT1 = """`allow_interpolation` was set to `False` in `Dataset` constructor. It means that either
1. there is a mistake in your code and you did not meant to perform this operation with an unaligned Footprint,
2. or that you want to perform a resampling operation and that you need `allow_interpolation` to be `True`.
"""

_EXN_FORMAT2 = """The `interpolation` parameter is None. It means that either
1. there is a mistake in your code and you did not meant to perform this operation with an unaligned Footprint,
2. or that you want to perform a resampling operation and that you need `interpolation` to be a string.
"""

class ABackSourceRasterRemapMixin(object):
    """Raster Mixin containing remap subroutine"""

    _REMAP_MASK_MODES = frozenset(['dilate', 'erode', ])
    REMAP_INTERPOLATIONS = {
        'cv_area': cv2.INTER_AREA,
        'cv_nearest': cv2.INTER_NEAREST,
        'cv_linear': cv2.INTER_LINEAR,
        'cv_cubic': cv2.INTER_CUBIC,
        'cv_lanczos4': cv2.INTER_LANCZOS4,
    }

    def build_sampling_footprint(self, fp, interpolation):
        if not fp.share_area(self.fp):
            return None
        if fp.same_grid(self.fp):
            fp = fp & self.fp
            assert fp.same_grid(self.fp)
            return fp
        return self.build_sampling_footprint_to_remap_interpolate(fp, interpolation)

    def build_sampling_footprint_to_remap_interpolate(self, fp, interpolation):
        if interpolation is None: # pragma: no cover
            raise ValueError(_EXN_FORMAT0.format(
                src=self.fp,
                dst=fp,
                tldiff=fp.tl - (
                    self.fp.pxtbvec * np.around(~self.fp.affine * fp.tl)[1] +
                    self.fp.pxlrvec * np.around(~self.fp.affine * fp.tl)[0]
                ) - self.fp.tl,
            ) + _EXN_FORMAT2)
        if not self.back_ds.allow_interpolation: # pragma: no cover
            raise ValueError(_EXN_FORMAT0.format(
                src=self.fp,
                dst=fp,
                tldiff=fp.tl - (
                    self.fp.pxtbvec * np.around(~self.fp.affine * fp.tl)[1] +
                    self.fp.pxlrvec * np.around(~self.fp.affine * fp.tl)[0]
                ) - self.fp.tl,
            ) + _EXN_FORMAT1)
        if interpolation in {'cv_nearest'}:
            dilate_size = 1 * self.fp.pxsizex / fp.pxsizex # hyperparameter
        elif interpolation in {'cv_linear', 'cv_area'}:
            dilate_size = 2 * self.fp.pxsizex / fp.pxsizex # hyperparameter
        else:
            dilate_size = 4 * self.fp.pxsizex / fp.pxsizex # hyperparameter
        dilate_size = max(2, np.around(dilate_size)) # hyperparameter too
        fp = fp.dilate(dilate_size)
        fp = self.fp & fp
        return fp

    @classmethod
    def remap(cls, src_fp, dst_fp, array, mask, src_nodata, dst_nodata, mask_mode, interpolation):
        """General remapping function from one Footprint to another.

        Caveat
        ------
        For performance reasons, output array might share the same memory space as the input array.
        If a nodata conversion is then performed, the input array will be modified.
        """
        # Parameters cheking ******************************************************************** **
        arr_mode = array is not None, mask is not None
        fp_mode = (
            src_fp.same_grid(dst_fp),
            src_fp.poly.contains(dst_fp.poly),
        )

        # Check array / mask ***************************************************
        if arr_mode[0]:
            assert array.shape[:2] == tuple(src_fp.shape)
            assert 2 <= array.ndim <= 3
            arr_outdtype = array.dtype
            arr_outshape = list(array.shape)
            arr_outshape[:2] = dst_fp.shape
            array = np.atleast_3d(array)
        if arr_mode[1]:
            assert mask.shape == tuple(src_fp.shape)
            assert mask.dtype == np.dtype('bool')

        # Check nodata *********************************************************
        assert dst_nodata is not None

        # Check mask_mode ******************************************************
        if mask_mode not in cls._REMAP_MASK_MODES: # pragma: no cover
            raise ValueError('mask_mode should be one of {}'.format(
                cls._REMAP_MASK_MODES,
            ))

        # Check interpolation **************************************************
        if not (interpolation is None or interpolation in cls.REMAP_INTERPOLATIONS): # pragma: no cover
            raise ValueError('interpolation should be None or one of {}'.format(
                cls.REMAP_INTERPOLATIONS.keys(),
            ))

        # Remapping ***************************************************************************** **
        if fp_mode == (True, True):
            array, mask = cls._remap_slice(
                src_fp, dst_fp,
                array, mask,
                src_nodata, dst_nodata,
            )
        elif fp_mode == (True, False):
            array, mask = cls._remap_copy(
                src_fp, dst_fp,
                array, mask,
                src_nodata, dst_nodata,
            )
        elif fp_mode == (False, ANY):
            array, mask = cls._remap_interpolate(
                src_fp, dst_fp,
                array, mask,
                src_nodata, dst_nodata,
                mask_mode, interpolation,
            )
        else:
            assert False # pragma: no cover

        # Return ******************************************************************************** **
        if arr_mode[0]:
            array = array.reshape(arr_outshape).astype(arr_outdtype, copy=False)

        if arr_mode == (True, True):
            return array, mask
        elif arr_mode == (True, False):
            return array
        elif arr_mode == (False, True):
            return mask
        else:
            assert False # pragma: no cover

    @staticmethod
    def _remap_slice(src_fp, dst_fp, array, mask, src_nodata, dst_nodata):
        src_slice = dst_fp.slice_in(src_fp)

        if array is not None:
            array = array[src_slice]
            if src_nodata is not None and dst_nodata != src_nodata:
                array[array == src_nodata] = dst_nodata
        if mask is not None:
            mask = mask[src_slice]
        return array, mask

    @staticmethod
    def _remap_copy(src_fp, dst_fp, array, mask, src_nodata, dst_nodata):
        dst_slice = src_fp.slice_in(dst_fp, clip=True)
        src_slice = dst_fp.slice_in(src_fp, clip=True)

        if array is not None:
            dstarray = np.full(np.r_[dst_fp.shape, array.shape[-1]], dst_nodata, array.dtype)
            dstarray[dst_slice] = array[src_slice]
            if src_nodata is not None and dst_nodata != src_nodata:
                dstarray[dst_slice][dstarray[dst_slice] == src_nodata] = dst_nodata
        else:
            dstarray = None # pragma: no cover

        if mask is not None:
            dstmask = np.full(dst_fp.shape, 0, mask.dtype)
            dstmask[dst_slice] = mask[src_slice]
        else:
            dstmask = None # pragma: no cover

        return dstarray, dstmask

    @classmethod
    def _remap_interpolate(cls, src_fp, dst_fp, array, mask, src_nodata, dst_nodata,
                           mask_mode, interpolation):
        if array is not None and array.dtype in [np.dtype('float64'), np.dtype('bool')]:
            raise ValueError(
                'dtype {!r} not handled by cv2.remap'.format(array.dtype)
            ) # pragma: no cover

        mapx, mapy = dst_fp.meshgrid_raster_in(src_fp, dtype='float32')
        mapx, mapy = cv2.convertMaps(
            mapx, mapy, cv2.CV_16SC2,
            nninterpolation=interpolation == 'cv_nearest',
        ) # At this point mapx/mapy are not really mapx/mapy any more, but who cares?
        interpolation = cls.REMAP_INTERPOLATIONS[interpolation]

        if array is not None:
            # "Bug" 1 with cv2.BORDER_CONSTANT *********************************
            # cv2.remap with cv2.BORDER_CONSTANT considers that the constant value is part of the
            # input values and interpolates with it. Which is obviously a problem...
            # If performing resampling for nodata and values separately, the result is safe with
            # cv2.BORDER_CONSTANT, since the pixels that have been wrongly interpolated with nodata
            # are overriden afterward.
            # "Bug" 2 with cv2.BORDER_CONSTANT *********************************
            # When building the multichannel nodata mask, only the first band works correctly near
            # borders
            # Solution *********************************************************
            # cv2.BORDER_TRANSPARENT + `dst` parameter is the safe way
            #
            #  "Bug" 3 with remap **********************************************
            # If the input array has shape (Y, X, 1), the output is (Y, X).
            if src_nodata is None:
                dstarray = np.full(np.r_[dst_fp.shape, array.shape[-1]], dst_nodata, array.dtype)
                cv2.remap(
                    array, mapx, mapy,
                    interpolation=interpolation,
                    borderMode=cv2.BORDER_TRANSPARENT,
                    dst=dstarray,
                )
            else:
                dstnodatamask = np.ones(np.r_[dst_fp.shape, array.shape[-1]], 'float32')
                cv2.remap(
                    (array == src_nodata).astype('float32', copy=False), mapx, mapy,
                    interpolation=interpolation,
                    dst=dstnodatamask,
                    borderMode=cv2.BORDER_TRANSPARENT,
                )
                dstnodatamask = dstnodatamask != 0
                dstarray = np.full(np.r_[dst_fp.shape, array.shape[-1]], dst_nodata, array.dtype)
                cv2.remap(
                    array, mapx, mapy,
                    interpolation=interpolation,
                    borderMode=cv2.BORDER_TRANSPARENT,
                    dst=dstarray,
                )
                dstarray[dstnodatamask] = dst_nodata
        else:
            dstarray = None # pragma: no cover

        if mask is not None:
            if mask_mode == 'erode':
                dstmask = cv2.remap(
                    mask.astype('float32', copy=False), mapx, mapy,
                    interpolation=interpolation,
                    borderMode=cv2.BORDER_CONSTANT,
                    borderValue=0.,
                ) == 1.
            elif mask_mode == 'dilate':
                dstmask = cv2.remap(
                    mask.astype('float32', copy=False), mapx, mapy,
                    interpolation=interpolation,
                    borderMode=cv2.BORDER_CONSTANT,
                    borderValue=0.,
                ) != 0.
            else:
                assert False # pragma: no cover
        else:
            dstmask = None # pragma: no cover

        return dstarray, dstmask
